import obspy
import os, glob
import time
import pyasdf
import sys
import pandas as pd
from obspy import read_events, UTCDateTime,read_inventory
from obspy.core import Trace, Stream
from obspy.clients.fdsn import Client
from seisgo import utils
from seisgo.utils import get_tracetag, save2asdf
import numpy as np

def get_sta_list(net_list, sta_list, chan_list, starttime, endtime, fname=None,\
                maxseischan=3,source='IRIS',lamin= None, lamax= None, \
                lomin= None, lomax= None, pressure_chan=None):
    """
    Function to get station list with given parameters. It is a wrapper of the obspy function "get_stations()".
    it is a practical applicaiton of get_stations() for mass downloading.

    ===PARAMETERS====
    net_list,sta_list,chan_list: targeted station information, wildcards are accepted.
    starttime,endtime: could be string (e.g., "2021_09_01_0_0_0") or obspy.UTCDateTime() objects.
    fname: save station list to file. Not save if None (default). Returns a Pandas dataframe of the list if None.
    maxseischan: default 3. this is used to avoid duplicates when, for example, both EH* and BH* channels are listed.
                The funciton will search by order in chan_list, and stop when maximum seismic channels are found.
                pressure channels are dealt with seperatedly.
    source: obspy Client source. Check obspy for the list of viable sources. Default is "IRIS". It can also be a Client object.
    lamin, lamax,lomin,lomax: box region for the search. optional.
    pressure_chan: list of pressure channels. This is used when counting the number of seismic channels. pressure channels
                are not counted. we could avoid duplicates but still keep the channels we want.
    """
    sta = [];
    net = [];
    chan = [];
    location = [];
    lon = [];
    lat = [];
    elev = []
    if isinstance(source,str):
        if source == 'IRISPH5':
            client=Client(service_mappings={'station':'http://service.iris.edu/ph5ws/station/1'})
        else:
            client=Client(source)
    elif isinstance(source,obspy.clients.fdsn.client.Client) or \
         isinstance(source,obspy.clients.fdsn.routing.federator_routing_client.FederatorRoutingClient) or \
         isinstance(source,obspy.clients.fdsn.routing.eidaws_routing_client.EIDAWSRoutingClient):
        client = source
    else:
        raise ValueError(str(type(source))+" is not supported.")

    # time tags
    if not isinstance(starttime,obspy.core.utcdatetime.UTCDateTime) and starttime is not None:
        starttime_UTC = obspy.UTCDateTime(starttime)
    else:
        starttime_UTC = starttime
    if not isinstance(endtime,obspy.core.utcdatetime.UTCDateTime)  and endtime is not None:
        endtime_UTC   = obspy.UTCDateTime(endtime)
    else:
        endtime_UTC = endtime
    # loop through specified network, station and channel lists
    chanhistory = {}

    if isinstance(net_list,str):net_list=[net_list]
    if isinstance(sta_list,str):sta_list=[sta_list]
    if isinstance(chan_list,str):chan_list=[chan_list]
    if isinstance(pressure_chan,str):pressure_chan=[pressure_chan]

    for inet in net_list:
        for ista in sta_list:
            dataflag = 0
            ista=str(ista)
            for ichan in chan_list:
                # gather station info
                try:
                    inv = client.get_stations(network=inet, station=ista, channel=ichan, location='*', \
                                              starttime=starttime_UTC, endtime=endtime_UTC, minlatitude=lamin,
                                              maxlatitude=lamax, \
                                              minlongitude=lomin, maxlongitude=lomax, level='response')
                    dataflag = 1
                except Exception as e:
                    if ichan == chan_list[-1] and dataflag == 0:
                        print(inet + '.' + ista + '.' + ichan + ': Abort due to ' + str(e))
                        break
                    else:
                        continue

                for K in inv:
                    for tsta in K:
                        # if ichan in pressure_chan or len(chanhistory)<maxseischan:
                        ckeys = chanhistory.keys()
                        netsta = K.code + '.' + tsta.code
                        for c in tsta.get_contents()['channels']:
                            chan_this=c.split('.')[-1]
                            if netsta not in ckeys:
                                chanhistory[netsta] = []
                                sta.append(tsta.code)
                                net.append(K.code)
                                chan.append(chan_this)
                                lon.append(tsta.longitude)
                                lat.append(tsta.latitude)
                                elev.append(tsta.elevation)
                                # sometimes one station has many locations and here we only get the first location
                                if tsta[0].location_code:
                                    location.append(tsta[0].location_code)
                                else:
                                    location.append('*')
                                if pressure_chan != None:
                                    if ichan not in pressure_chan:
                                        chanhistory[netsta].append(chan_this)
                                else:
                                    chanhistory[netsta].append(chan_this)
                            elif len(chanhistory[netsta]) < maxseischan or (pressure_chan !=None and chan_this in pressure_chan):
                                sta.append(tsta.code)
                                net.append(K.code)
                                chan.append(chan_this)
                                lon.append(tsta.longitude)
                                lat.append(tsta.latitude)
                                elev.append(tsta.elevation)
                                # sometimes one station has many locations and here we only get the first location
                                if tsta[0].location_code:
                                    location.append(tsta[0].location_code)
                                else:
                                    location.append('*')
                                if pressure_chan != None:
                                    if ichan not in pressure_chan:
                                        chanhistory[netsta].append(chan_this)
                                else:
                                    chanhistory[netsta].append(chan_this)

    # output station list
    dict = {'network': net, 'station': sta, 'channel': chan, 'latitude': lat, 'longitude': lon, 'elevation': elev}
    locs0 = pd.DataFrame(dict)
    locs=locs0.drop_duplicates(ignore_index=True) #remove duplicate inputs/stations
    if fname is not None:
        fhead,ftail=os.path.split(fname)
        if len(fhead)>0:
            if not os.path.isdir(fhead):os.makedirs(fhead)
        locs.to_csv(fname, index=False)

    return locs

#
def getdata(net,sta,starttime,endtime,chan,source='IRIS',samp_freq=None,
            rmresp=True,rmresp_output='VEL',pre_filt=None,debug=False,
            sacheader=False,getstainv=False,verbose=False):
    """
    This is a wrapper that downloads seismic data and (optionally) removes response
    and downsamples if needed. Most of the arguments have the same meaning as for
    obspy.Client.get_waveforms().

    Parameters
    ----------
    net,sta,chan : string
            network, station, and channel names for the request.
    starttime, endtime : UTCDateTime
            Starting and ending date time for the request.
    source : string.  It can also be a Client object.
            Client names.
            To get a list of available clients:
            >> from obspy.clients.fdsn.header import URL_MAPPINGS
            >> for key in sorted(URL_MAPPINGS.keys()):
                 print("{0:<11} {1}".format(key,  URL_MAPPINGS[key]))
    samp_freq : float
            Target sampling rate. Skip resampling if None.
    rmresp : bool
            Remove response if true. For the purpose of download OBS data and remove
            tilt and compliance noise, the output is "VEL" for pressure data and "DISP"
            for seismic channels.
    rmresp_output : string
            Output format when removing the response, following the same rule as by OBSPY.
            The default is 'VEL' for velocity output.
    pre_filt : :class: `numpy.ndarray`
            Same as the pre_filt in obspy when removing instrument responses.
    debug : bool
            Plot raw waveforms before and after preprocessing.
    sacheader : bool
            Key sacheader information in a dictionary using the SAC header naming convention.
    """
    if isinstance(source,str):
        if source == 'IRISPH5':
            client=Client(service_mappings={'dataselect':'http://service.iris.edu/ph5ws/dataselect/1',
                        'station':'http://service.iris.edu/ph5ws/station/1'})
        else:
            client=Client(source)
    elif isinstance(source,obspy.clients.fdsn.client.Client) or \
         isinstance(source,obspy.clients.fdsn.routing.federator_routing_client.FederatorRoutingClient) or \
         isinstance(source,obspy.clients.fdsn.routing.eidaws_routing_client.EIDAWSRoutingClient):
        client = source
    else:
        raise ValueError(str(type(source))+" is not supported.")

    tr = None
    sac=dict() #place holder to save some sac headers.
    #check arguments
    if rmresp:
        if pre_filt is None:
            pre_filt = set_filter(samp_freq, 0.001)
            print("getdata(): pre_filt is not specified. Use 0.001-0.5*samp_freq as "+\
                    "the filter range when removing response.")
    """
    a. Downloading
    """
    if sacheader or getstainv or rmresp:
        inv = client.get_stations(network=net,station=sta,
                        channel=chan,location="*",starttime=starttime,endtime=endtime,
                        level='response')
        if sacheader:
            tempsta,tempnet,stlo, stla,stel,temploc=utils.sta_info_from_inv(inv)
            sac['knetwk']=tempnet
            sac['kstnm']=tempsta
            sac['stlo']=stlo
            sac['stla']=stla
            sac['stel']=stel
            sac['kcmpnm']=chan
            sac['khole']=temploc

    # pressure channel
    tr=client.get_waveforms(network=net,station=sta,
                    channel=chan,location="*",starttime=starttime,endtime=endtime)
#     trP[0].detrend()
    if verbose:print('number of segments downloaded: '+str(len(tr)))

    tr[0].stats['sac']=sac

    if verbose:print("station "+net+"."+sta+" --> seismic channel: "+chan)

    if debug:
        year = tr[0].stats.starttime.year
        julday = tr[0].stats.starttime.julday
        hour = tr[0].stats.starttime.hour
        mnt = tr[0].stats.starttime.minute
        sec = tr[0].stats.starttime.second
        tstamp = str(year) + '.' + str(julday)+'T'+str(hour)+'-'+str(mnt)+'-'+str(sec)
        trlabels=[net+"."+sta+"."+tr[0].stats.channel]
    """
    b. Resampling
    """
    if samp_freq is not None:
        sps=int(tr[0].stats.sampling_rate)
        delta = tr[0].stats.delta
        #assume pressure and vertical channels have the same sampling rat
        # make downsampling if needed
        if sps > samp_freq:
            if verbose:print("  downsamping from "+str(sps)+" to "+str(samp_freq))
            for r in tr:
                if np.sum(np.isnan(r.data))>0:
                    raise(Exception('NaN found in trace'))
                else:
                    r.interpolate(samp_freq,method='weighted_average_slopes')
                    # when starttimes are between sampling points
                    fric = r.stats.starttime.microsecond%(delta*1E6)
                    if fric>1E-4:
                        r.data = utils.segment_interpolate(np.float32(r.data),float(fric/(delta*1E6)))
                        #--reset the time to remove the discrepancy---
                        r.stats.starttime-=(fric*1E-6)
                # print('new sampling rate:'+str(tr.stats.sampling_rate))

    """
    c. Plot raw data before removing responses.
    """
    if debug:
        utils.plot_trace([tr],size=(12,3),title=trlabels,freq=[0.005,0.1],ylabels=["raw"],
                        outfile=net+"."+sta+"_"+tstamp+"_raw.png")

    """
    d. Remove responses
    """
    if rmresp:
        for r in tr:
            if np.sum(np.isnan(r.data))>0:
                raise(Exception('NaN found in trace'))
            else:
                try:
                    if verbose:print('  removing response using inv for '+net+"."+sta+"."+r.stats.channel)
                    r.attach_response(inv)
                    r.remove_response(output=rmresp_output,pre_filt=pre_filt,
                                              water_level=60,zero_mean=True,plot=False)
                except Exception as e:
                    print(e)
                    r = []
    if verbose:print('cleaning with demean, detrend, and taper')
    for r in tr:
        if verbose:print('    --> demean')
        r.detrend('demean')
        if verbose:print('    --> detrend')
        r.detrend('linear')
        if verbose:print('    --> taper')
        r.data = utils.taper(r.data, fraction=0.005)
        # r.taper(0.005)

    if len(tr.get_gaps())>0:
        if verbose:print('merging segments with gaps')
        tr.merge(fill_value=0)
    tr=tr[0]
    """
    e. Plot raw data after removing responses.
    """
    if debug:
        plot_trace([tr],size=(12,3),title=trlabels,freq=[0.005,0.1],ylabels=[rmresp_output],
                   outfile=net+"."+sta+"_"+tstamp+"_raw_rmresp.png")

    #
    if getstainv:return tr,inv
    else: return tr

def cleantargetdir(rawdatadir):
    if not os.path.isdir(rawdatadir):
        print("target directory not found")
    dfiles1 = glob.glob(os.path.join(rawdatadir, '*.h5'))
    if len(dfiles1) > 0:
        print('Cleaning up raw data directory before downloading ...')
        for df1 in dfiles1:
            os.remove(df1)

def in_directory(fname, sta, net, tag):
    # Returns False if not already in directory
    if not os.path.isfile(fname):
        return False
    else:
        # Checks if file w/ same channel in directory
        with pyasdf.ASDFDataSet(fname, mpi=False, mode='r') as rds:
            read_tnames = rds.waveforms.list()
            read_tags= [sta.get_waveform_tags() for sta in rds.waveforms]
            read_records = {read_tnames[i]: read_tags[i][0] for i in range(len(read_tnames))}
            tname = net + '.' + sta
            # print(read_records)
            # if tname in read_records and tag == tname[tag]:
            # if {tname:tag}.items() <= read_records.items():
            if ((tname, tag) in read_records.items()):
                return True
            else:
                return False

def set_filter(samp_freq, pfreqmin,pfreqmax=None):
    if pfreqmax is None:
        pfreqmax = samp_freq / 2
    f1 = 0.95 * pfreqmin;
    f2 = pfreqmin
    if 1.05 * pfreqmax > 0.48 * samp_freq:
        f3 = 0.45 * samp_freq
        f4 = 0.48 * samp_freq
    else:
        f3 = pfreqmax
        f4 = 1.05 * pfreqmax
    pre_filt = [f1, f2, f3, f4]
    return pre_filt

def download(starttime, endtime, stationinfo=None, network=None, station=None,channel=None,
                source='IRIS',rawdatadir=None,sacheader=False, getstainv=True, max_tries=10,
                savetofile=False,pressure_chan=None,samp_freq=None,freqmin=0.001,freqmax=None,
                rmresp=True, rmresp_out='DISP',respdir=None,qc=True,event=None,verbose=False):
    """
    starttime, endtime: timing duration for the download.
    stationinfo:
            This could be a CSV file name, a Pandas DataFrame, or a dictionary. All data should
            include network, station, (optional) channel, (optional) location. Note, if stationinfo is NOT None,
            the station information in stationinfo will overwrite the network,station, and channel,
            if specified individually.
    network,station,channel: those will be ignored if stationinfo is NOT None.
    source: obspy source name. default is "IRIS"
    rawdatadir: directory to save the downloaded data. only needed when savetofile is True.
    sacheader: default False. get sacheader or not. this is useful when you want to save data to sac later or check some metadata.
    getstainv: default True. get station inventory. If true, funciton returns two variables: trlist,invlist
    max_tries: defautl 10. will try multiple times incase there is server connection issue.
    savetofile: default is False. Save to ASDF file if True.
    pressure_chan: this is needed when downloading OBS data with pressure channel. this influences the output of
            response removal. Use "PRESSURE" for pressure channels to get comparable amplitudes as "DISP" for seismic channels.
    samp_freq: downsample data to target sampling rate.
    freqmin,freqmax: frequency range in removing response. default freqmin is 0.001.
    rmresp: default True. remove response.
    rmresp_out: default "DISP". use "PRESSURE" for pressure chan.
    respdir: directory containing response information. currently NOT used in this function.
    qc: When True, does QC to clean up the trace. default True.
    event: ObsPy Event object for earthquake data. the event qml data will be saved into ASDF.
    verbose: verbose flag. print some intermediate steps/info if True. Default False.
    =============RETURNS============
    trlist: Obspy Stream containing all traces. Note that when savetofile is True, the return will be an empty Stream.
    sta_inv_list: inventory list of the stations. Empty when getstainv is False.
    """
    ######################read in station information first.
    if stationinfo is not None:
        if isinstance(stationinfo,str): #assume stationinfo is a CSV file containing network,station, channel columns
            if not os.path.isfile(stationinfo):raise IOError('file %s not exist! double check!' % stationinfo)
            # read station info from list
            locs = pd.read_csv(stationinfo)
        elif isinstance(stationinfo,pd.DataFrame) or isinstance(stationinfo,dict):
            locs = stationinfo

        network  = list(locs['network'])
        station  = list(locs['station'])
        try:
            channel = list(locs['channel'])
        except Exception as e:
            channel = ['*']*len(station)
        # location info: useful for some occasion
        try:
            location = list(locs['location'])
        except Exception as e:
            location = ['*']*len(station)
    elif None in [network,station,channel]:
        raise IOError('Must specify network, station, and channel if stationinfo is None!')
    ##########################################
    if rawdatadir is not None:
        savetofile=True
        if not os.path.isdir(rawdatadir):os.makedirs(rawdatadir)

    if event is None:
        type = "continuous"
    else:
        if event.event_type == "earthquake":
            type = "earthquake"
        else:
            type = "other event"

    # if user passes a string instead of a list, make a list of one string
    # if station is None: station = ['*']*len(network)
    if isinstance(station, str): station = [station]
    # if channel is None: channel = ['*']*len(station)
    if isinstance(channel, str): channel = [channel]
    if isinstance(network, str): network = [network]
    if isinstance(pressure_chan,str):pressure_chan=[pressure_chan]

    pre_filt = set_filter(samp_freq, freqmin,freqmax)

    # dtlist = utils.split_datetimestr(starttime, endtime, inc_hours)
    # print(dtlist)
    # for idt in range(len(dtlist) - 1):
    if isinstance(starttime,str):
        sdatetime = obspy.UTCDateTime(starttime)
    else:
        sdatetime = starttime
    if isinstance(endtime,str):
        edatetime = obspy.UTCDateTime(endtime)
    else:
        edatetime = endtime

    if savetofile:
        if type == "continuous":
            fname = os.path.join(rawdatadir,
                                 str(sdatetime).replace(':', '-') + 'T' + str(edatetime).replace(':', '-') + '.h5')
        elif type == 'earthquake':
            fname = os.path.join(rawdatadir, str(event.origins[0].time) + "_M" + str(event.magnitudes[0].mag) + '.h5')
        else:
            fname = os.path.join(rawdatadir, str(event.origins[0].time) + '.h5')

    """
    Start downloading.
    """
    trlist=[]
    sta_inv_list=[]
    #loop through all stations.
    for i in range(len(station)):
        inet=network[i]
        ista=str(station[i])
        ichan=channel[i]

        for nt in range(max_tries):
            if verbose:print(inet+'.'+ista + '.' + ichan + '  downloading ... try ' + str(nt + 1))

            t0 = time.time()

            rmresp_out_tmp=rmresp_out
            if pressure_chan is not None and ichan in pressure_chan:
                rmresp_out_tmp='PRESSURE'
            try:
                output = getdata(inet, ista, sdatetime, edatetime, chan=ichan, source=source,
                                        samp_freq=samp_freq, rmresp=rmresp, rmresp_output=rmresp_out_tmp,
                                       pre_filt=pre_filt, sacheader=sacheader, getstainv=getstainv,verbose=verbose)
            except Exception as e:
                print(e, 'for', ista)
                time.sleep(1)  # sleep for 1 second before next try.
                continue
            if getstainv == True or sacheader == True:
                sta_inv = output[1]
                tr = output[0]
            else:
                tr = output
                sta_inv = None

            ta = time.time() - t0
            if verbose:print('  downloaded ' + inet+"." + ista + "." + ichan + " in " + str(ta) + " seconds.")
            tag = get_tracetag(tr)
            chan = tr.stats.channel

            """
            Add cleanup
            """
            if qc:
                if chan[-1].lower() == 'h': tag_type= "trP"; hasPressure=True
                elif chan[-1].lower() == '1' or chan[-1].lower() == 'e':tag_type="tr1"
                elif chan[-1].lower() == '2' or chan[-1].lower() == 'n':tag_type="tr2"
                elif chan[-1].lower() == 'z':tag_type="trZ"
                else: print('  No seismic channels found. Drop the station: '+ista); break

                #sanity check.
                badtrace=False
                if not isinstance(tr, Trace):
                    print("  "+str(tr)+" is not a Trace object. "+ista)
                    badtrace=True
                    break
                elif np.sum(np.isnan(tr.data))>0:
                    print('  NaN found in trace: '+str(tr)+". "+ista)
                    badtrace=True
                    break
                elif np.count_nonzero(tr.data) < 1:
                    print('  All zeros in trace: '+str(tr)+". "+ista)
                    badtrace=True
                    break

                if badtrace:
                    if savetofile:
                        in_dir = in_directory(fname, ista, inet, tag)
                        """
                        Save to ASDF file.
                        """
                        if in_dir == True:
                            print(inet + '.' + ista + '.' + chan + '  exists. Continue to next!')
                            break
                        else:
                            print("  Encountered bad trace for " + ista + ". Save as is without processing!")
                            save2asdf(fname,tr,tag,sta_inv=sta_inv, event=event)
                            break
                    else:
                        trlist.append(tr)
                        sta_inv_list.append(sta_inv)
                        break
                else:
                    if savetofile:
                        in_dir = in_directory(fname, ista, inet, tag)
                        """
                        Save to ASDF file.
                        """
                        if in_dir == True:
                            print(inet + '.' + ista + "." + chan + '  exists. Continue to next!')
                            break
                        else:
                            print(" Saving data for", inet + '.' + ista  + '.' + chan )
                            save2asdf(fname,[tr],[tag],sta_inv=sta_inv, event=event)
                            break
                    else:
                        trlist.append(tr)
                        sta_inv_list.append(sta_inv)
                        break
            else:  #not QC
                if savetofile:
                    print(" Saving data for", inet + '.' + ista  + '.' + chan )
                    save2asdf(fname,[tr],[tag],sta_inv=sta_inv, event=event)
                    break
                else:
                    trlist.append(tr)
                    sta_inv_list.append(sta_inv)
                    break

    return Stream(trlist),sta_inv_list

def read_data(files,rm_resp='no',respdir='.',freqmin=None,freqmax=None,rm_resp_out='VEL',
                getstainv=False,water_level=60,samp_freq=None,stainfo=None):
    """
    Wrapper to read local data and (optionally) remove instrument response, and gather station inventory.

    ==== PARAMETERS ====
    files: local data file or a list of files.
    rm_resp: 'no'[default], 'spectrum', 'RESP', 'polozeros', or 'inv'.
    respdir: directory for response files, default is '.'
    freqmin: minimum frequency in removing responses. default is 0.001
    freqmax: maximum frequency in removing responses. default is 0.499*sample_rate
    rm_resp_out: the ouptut unit for removing response, default is 'VEL', could be "DIS"
    getstainv: get station inventory or not, default is False.
    stainfo:  pandas dataframe contaning station information: network,station,latitudie,longitude,elevation

    ==== RETURNS ====
    tr_all: all traces as a list of obspy Trace objects.
    inv_all: all inventory objects, if stainv flag is True. Otherwise, only tr_all will be returned.
    """
    if isinstance(files,str):files=[files]
    tr_all=[]
    inv_all=[]
    for f in files:
        tr=obspy.read(f, debug_headers=True)

        fs=tr[0].stats.sampling_rate
        if freqmin is None: freqmin=0.001
        if freqmax is None: freqmax=0.499*fs
        pre_filt = set_filter(fs, freqmin,freqmax)
        net=tr[0].stats.network
        sta=tr[0].stats.station
        loc=tr[0].stats.location
        chan=tr[0].stats.channel
        netstachan=net+"."+sta+"."+str(loc)+"."+chan
        date_info = {'starttime':tr[0].stats.starttime,'endtime':tr[0].stats.endtime}
        if rm_resp != 'no':
            if rm_resp == 'spectrum':
                print('remove response using spectrum')
                specfile = glob.glob(os.path.join(respdir,'*'+netstachan+'*'))
                if len(specfile)==0:
                    raise ValueError('no response sepctrum found for %s' % netstachan)
                tr = utils.resp_spectrum(tr,specfile[0],fs,pre_filt)

            elif rm_resp == 'RESP':
                print('remove response using RESP files')
                resp = glob.glob(os.path.join(respdir,'RESP.'+netstachan+'*'))
                print(resp)
                if len(resp)==0:
                    raise ValueError('no RESP files found for %s' % netstachan)
                seedresp = {'filename':resp[0],'date':date_info['starttime'],'units':rm_resp_out}
                tr.simulate(paz_remove=None,pre_filt=pre_filt,seedresp=seedresp)

            elif rm_resp == 'polozeros':
                print('remove response using polos and zeros')
                paz_sts = glob.glob(os.path.join(respdir,'*'+netstachan+'*'))
                if len(paz_sts)==0:
                    raise ValueError('no polozeros found for %s' % netstachan)
                tr.simulate(paz_remove=paz_sts[0],pre_filt=pre_filt)
            elif rm_resp == 'inv':
                print('remove response using inventory')
                inv = read_inventory(os.path.join(respdir,'*'+net+"."+sta+'*.xml'))
                tr.remove_response(inventory=inv,pre_filt=pre_filt,output=rm_resp_out,water_level=water_level)
            else:
                raise ValueError('no such option for rm_resp! please double check!')

        if samp_freq is not None:
            sps=int(tr[0].stats.sampling_rate)
            delta = tr[0].stats.delta
            #assume pressure and vertical channels have the same sampling rat
            # make downsampling if needed
            if sps > samp_freq:
                print("  downsamping from "+str(sps)+" to "+str(samp_freq))
                if np.sum(np.isnan(tr[0].data))>0:
                    raise(Exception('NaN found in trace'))
                else:
                    tr[0].interpolate(samp_freq,method='weighted_average_slopes')
                    # when starttimes are between sampling points
                    fric = tr[0].stats.starttime.microsecond%(delta*1E6)
                    if fric>1E-4:
                        tr[0].data = utils.segment_interpolate(np.float32(tr[0].data),float(fric/(delta*1E6)))
                        #--reset the time to remove the discrepancy---
                        tr[0].stats.starttime-=(fric*1E-6)
        tr[0].detrend('demean')
        tr[0].detrend('linear')
        tr[0].taper(0.005)
        tr_all.append(tr[0])

        if getstainv:
            if isinstance(stainfo,pd.DataFrame) or isinstance(stainfo,dict) or stainfo is None:
                inv_all.append(utils.stats2inv(tr[0].stats,locs=stainfo))
            else:
                inv_all.append(read_inventory(stainfo))
    #
    #return
    if getstainv:
        return Stream(tr_all),inv_all
    else:
        return Stream(tr_all)

def ms2asdf(files,rm_resp='no',respdir='.',respfile=None,freqmin=None,freqmax=None,rm_resp_out='VEL',
                water_level=60,samp_freq=None,outdir='.',stainfo=None,outfile=None):
    """
    Wrapper to read local data and (optionally) remove instrument response, and gather station inventory.

    ==== PARAMETERS ====
    files: local data file or a list of files.
    rm_resp: 'no'[default], 'spectrum', 'RESP', 'polozeros', or 'inv'.
    respdir: directory for response files, default is '.'
    respfile: response file or station xml. default None, looking for NET.STA.xml
    freqmin: minimum frequency in removing responses. default is 0.001
    freqmax: maximum frequency in removing responses. default is 0.499*sample_rate
    rm_resp_out: the ouptut unit for removing response, default is 'VEL', could be "DIS"
    outdir: directory to save the asdf file.
    outfile: filename

    """
    if isinstance(files,str):files=[files]
    #get some common parameters first:
    tr=obspy.read(files[0], debug_headers=True)

    fs=tr[0].stats.sampling_rate
    if freqmin is None: freqmin=0.001
    if freqmax is None: freqmax=0.499*fs
    pre_filt = set_filter(fs, freqmin,freqmax)
    inv=None
    if not os.path.isdir(outdir):os.makedirs(outdir)
    for f in files:
        tr=obspy.read(f, debug_headers=True)
        net=tr[0].stats.network
        sta=tr[0].stats.station
        chan=tr[0].stats.channel
        netstachan=net+"."+sta+"."+chan
        date_info = {'starttime':tr[0].stats.starttime,'endtime':tr[0].stats.endtime}
        if rm_resp != 'no':
            if rm_resp == 'spectrum':
                print('remove response using spectrum')
                if respfile is not None:
                    specfile=[respfile]
                else:
                    specfile = glob.glob(os.path.join(respdir,'*'+netstachan+'*'))
                if len(specfile)==0:
                    raise ValueError('no response sepctrum found for %s' % netstachan)
                tr = utils.resp_spectrum(tr,specfile[0],fs,pre_filt)

            elif rm_resp == 'RESP':
                print('remove response using RESP files')
                if respfile is None:
                    resp = glob.glob(os.path.join(respdir,'RESP.'+netstachan+'*'))
                else:
                    resp = [respfile]
                print(resp)
                if len(resp)==0:
                    raise ValueError('no RESP files found for %s' % netstachan)
                seedresp = {'filename':resp[0],'date':date_info['starttime'],'units':rm_resp_out}
                tr.simulate(paz_remove=None,pre_filt=pre_filt,seedresp=seedresp)

            elif rm_resp == 'polozeros':
                print('remove response using polos and zeros')
                if respfile is not None:
                    paz_sts = [respfile]
                else:
                    paz_sts = glob.glob(os.path.join(respdir,'*'+netstachan+'*'))
                if len(paz_sts)==0:
                    raise ValueError('no polozeros found for %s' % netstachan)
                tr.simulate(paz_remove=paz_sts[0],pre_filt=pre_filt)
            elif rm_resp == 'inv':
                print('remove response using inventory')
                if respfile is not None:
                    invfile=respfile
                else:
                    invfile='*'+net+"."+sta+'*.xml'
                inv = read_inventory(os.path.join(respdir,invfile))
                tr.remove_response(inventory=inv,pre_filt=pre_filt,output=rm_resp_out,water_level=water_level)
            else:
                raise ValueError('no such option for rm_resp! please double check!')

        if samp_freq is not None:
            sps=int(tr[0].stats.sampling_rate)
            delta = tr[0].stats.delta
            #assume pressure and vertical channels have the same sampling rat
            # make downsampling if needed
            if sps > samp_freq:
                print("  downsamping from "+str(sps)+" to "+str(samp_freq))
                if np.sum(np.isnan(tr[0].data))>0:
                    raise(Exception('NaN found in trace'))
                else:
                    tr[0].interpolate(samp_freq,method='weighted_average_slopes')
                    # when starttimes are between sampling points
                    fric = tr[0].stats.starttime.microsecond%(delta*1E6)
                    if fric>1E-4:
                        tr[0].data = utils.segment_interpolate(np.float32(tr[0].data),float(fric/(delta*1E6)))
                        #--reset the time to remove the discrepancy---
                        tr[0].stats.starttime-=(fric*1E-6)
        tr[0].detrend('demean')
        tr[0].detrend('linear')
        tr[0].taper(0.005)
        if rm_resp!='inv' and stainfo is not None:
            if isinstance(stainfo,pd.DataFrame) or isinstance(stainfo,dict):
                inv=utils.stats2inv(tr[0].stats,locs=stainfo)
            else:
                inv=read_inventory(stainfo)

        tag = get_tracetag(tr[0])
        sdatetime=tr[0].stats.starttime
        edatetime=tr[0].stats.endtime
        fname = os.path.join(outdir,str(sdatetime).replace(':', '-') + 'T' + str(edatetime).replace(':', '-') + '.h5')
        print(" Saving data to ", fname )
        utils.save2asdf(fname,[tr[0]],[tag],sta_inv=inv)

def get_events(start,end,minlon=-180,maxlon=180,minlat=-90,maxlat=90,minmag=0,maxmag=10,
                    magstep=1.0,mindepth=-100,maxdepth=1000,lat=None,lon=None,
                    minradius=None,maxradius=None,maxradiuskm=None,source="USGS",v=False):
    """
    Download event catalog within a box from USGS or ISC catalogs.
    """
    #elist is a list of panda dataframes
    t0=time.time()
    if None in [lat, lon, maxradius] and None in [lat, lon, maxradiuskm]:
        searchtype='box'
    else:
        searchtype='circle'
    if maxradius is not None and maxradiuskm is not None:
        raise ValueError('Only one of maxradius and maxradiuskm could be specified!'+\
                         " See https://earthquake.usgs.gov/fdsnws/event/1 for details.")
    
    if maxradiuskm is not None and source=="ISC":
        raise ValueError("maxrasiumkm is only used by USGS API.")
    if minradius is not None and source=="USGS":
        raise ValueError("minrasium is only used by ISC API. ")
    if minradius is None:
        minradius=0
    #Download the catalog by magnitudes
    maglist=np.arange(minmag,maxmag+0.5*magstep,magstep)
    events=[]
    for i in range(len(maglist)-1):
        if i>0:
            minM=str(maglist[i]+0.0001) #to avoid duplicates with intersecting magnitude thresholds
        else:
            minM=str(maglist[i])
        maxM=str(maglist[i+1])
        if v: print(minM,maxM)
        if searchtype == 'box':
            if source=="ISC":
                quake_url="http://isc-mirror.iris.washington.edu/fdsnws/event/1/query?starttime="+\
                start+"&endtime="+end+"&minlatitude="+str(minlat)+"&maxlatitude="+str(maxlat)+"&minlongitude="+\
                str(minlon)+"&maxlongitude="+str(maxlon)+"&minmagnitude="+minM+"&maxmagnitude="+maxM+\
                "&mindepth="+str(mindepth)+"&maxdepth="+str(maxdepth)+""
            elif source=="USGS":
                quake_url="https://earthquake.usgs.gov/fdsnws/event/1/query?format=xml&starttime="+\
                start+"&endtime="+end+"&minmagnitude="+minM+"&maxmagnitude="+maxM+"&minlatitude="+\
                str(minlat)+"&maxlatitude="+str(maxlat)+"&minlongitude="+str(minlon)+"&maxlongitude="+str(maxlon)+\
                "&mindepth="+str(mindepth)+"&maxdepth="+str(maxdepth)+""
        else:
            if source=="ISC":
                quake_url="http://isc-mirror.iris.washington.edu/fdsnws/event/1/query?starttime="+\
                start+"&endtime="+end+"&latitude="+str(lat)+"&longitude="+\
                str(lon)+"&minradius="+str(minradius)+"&maxradius="+str(maxradius)+"&minmagnitude="+minM+"&maxmagnitude="+maxM+\
                "&mindepth="+str(mindepth)+"&maxdepth="+str(maxdepth)+""
            elif source=="USGS":
                if maxradius is not None:
                    quake_url="https://earthquake.usgs.gov/fdsnws/event/1/query?format=xml&starttime="+\
                    start+"&endtime="+end+"&minmagnitude="+minM+"&maxmagnitude="+maxM+"&latitude="+\
                    str(lat)+"&longitude="+str(lon)+"&maxradius="+str(maxradius)+\
                    "&mindepth="+str(mindepth)+"&maxdepth="+str(maxdepth)+""
                else:
                    quake_url="https://earthquake.usgs.gov/fdsnws/event/1/query?format=xml&starttime="+\
                    start+"&endtime="+end+"&minmagnitude="+minM+"&maxmagnitude="+maxM+"&latitude="+\
                    str(lat)+"&longitude="+str(lon)+"&maxradiuskm="+str(maxradiuskm)+\
                    "&mindepth="+str(mindepth)+"&maxdepth="+str(maxdepth)+""

        try:
            event = read_events(quake_url)
            if v: print("Catalog: " + str(event.count()))
            events.append(event)
        except Exception as e:
            print(e)

    catalog = events[0]
    for i in np.arange(1,len(events)):
            catalog.extend(events[i])
    #Print the data frame, display total, and total time to complete task
    if v: print("It took "+str(time.time()-t0)+" seconds to download the catalog")

    return catalog

def get_event_waveforms(event,stainfo,window=150,offset=-50,arrival_type='first',maxtry=3,source='IRIS',rawdatadir=None,
                        sacheader=False, getstainv=False, savetofile=False,pressure_chan=None,
                        samp_freq=None,freqmin=0.001,freqmax=None,rmresp=True, rmresp_out='DISP',
                        respdir=None,qc=True):
    """
    Download seismic event waveforms. This function downloads one earthquake for one station.

    ====PARAMETERS====
    event: Obspy Event object.
    stainfo: Pandas dateframe contaning station information.
    window: length of the waveform in seconds
    offset: offset in seconds relative to the specified arrival.
    arrival_type: default is first arrival time, could be any specified phase string.
    other options are the same as download().

    savetofile: return Obspy Stream if False.
    """
    event_t = event.origins[0].time
    event_long = event.origins[0].longitude
    event_lat = event.origins[0].latitude
    depth_km = event.origins[0].depth / 1000
    network=list(stainfo.network)
    station=list(stainfo.station)
    try:
        channel = list(stainfo.channel)
    except Exception as e:
        channel = ['*']*len(station)
    trall=[]
    invall=[]
    # Download for each station
    for i in range(len(network)):
        sta_lat = stainfo.latitude[i]
        sta_long = stainfo.longitude[i]

        travel_time = utils.get_tt(event_lat, event_long, sta_lat, sta_long, depth_km,type='first')[0]
        sta_start = event_t + travel_time + offset
        sta_end = sta_start + window
        if getstainv:
            tr,inv=download(sta_start,sta_end,network=network[i],station=station[i],channel=channel[i],
                                source=source,rawdatadir=rawdatadir,sacheader=sacheader,
                                getstainv=getstainv, max_tries=maxtry,savetofile=savetofile,
                                pressure_chan=pressure_chan,samp_freq=samp_freq,freqmin=freqmin,
                                freqmax=freqmax,rmresp=rmresp, rmresp_out=rmresp_out,
                                respdir=respdir,qc=qc,event=event)

            if not savetofile:
                if len(tr)>0:
                    trall.append(tr[0])
                    invall.append(inv)
        else:
            tr=download(sta_start,sta_end,network=network[i],station=station[i],channel=channel[i],
                                source=source,rawdatadir=rawdatadir,sacheader=sacheader,
                                getstainv=getstainv, max_tries=maxtry,savetofile=savetofile,
                                pressure_chan=pressure_chan,samp_freq=samp_freq,freqmin=freqmin,
                                freqmax=freqmax,rmresp=rmresp, rmresp_out=rmresp_out,
                                respdir=respdir,qc=qc,event=event)

            if not savetofile:
                if len(tr)>0:
                    trall.append(tr[0])

    ###
    ###
    if not savetofile:
        if getstainv:
            return Stream(trall),invall
        else:
            return Stream(trall)



#!/usr/bin/env python
# coding: utf-8
"""
Download OBS date including all four components (seismic channels and the pressure channel).
This light-weight script mainly aims at downloading test data for some seisgo modules, instead of a comprehensive
and robust downloading wrapper. It is recommended to use NoisePy's downloading script for
more comprehensive downloading (https://github.com/mdenolle/NoisePy).
"""
#import needed packages.
from seisgo import utils
from seisgo import obsmaster as obs
import sys
import time
import scipy
import obspy
import pyasdf
import datetime
import os, glob
import numpy as np
import pandas as pd
import matplotlib.pyplot  as plt
from obspy import UTCDateTime
from obspy.core import Stream,Trace
from obspy.clients.fdsn import Client

# In[ ]:
"""
1. Set parameters for downloading data.
"""
rawdatadir = '../data/raw'
if not os.path.isdir(rawdatadir): os.mkdir(rawdatadir)

source='IRIS'
client=Client(source)
# get data from IRIS web service
net="7D"
stalist=["FN07A","G30A"]#["G03A","J35A","J44A","J65A"]

starttime = "2012_02_02_0_0_0"
endtime   = "2012_02_05_0_0_0"
inc_hours = 8

"""
2. Preprocessing parameters
"""
rmresp=True #remove instrument response
# parameters for butterworth filter
samp_freq=10
pfreqmin=0.002
pfreqmax=samp_freq/2

# prefilter information used when removing instrument responses
f1 = 0.95*pfreqmin;f2=pfreqmin
if 1.05*pfreqmax > 0.48*samp_freq:
    f3 = 0.45*samp_freq
    f4 = 0.48*samp_freq
else:
    f3 = pfreqmax
    f4= 1.05*pfreqmax
pre_filt  = [f1,f2,f3,f4]


"""
3. Download by looping through datetime list.
***** The users usually don't need to chance the following lines ****
"""
dtlist = utils.split_datetimestr(starttime,endtime,inc_hours)
print(dtlist)
for idt in range(len(dtlist)-1):
    sdatetime = obspy.UTCDateTime(dtlist[idt])
    edatetime   = obspy.UTCDateTime(dtlist[idt+1])

    fname = os.path.join(rawdatadir,dtlist[idt]+'T'+dtlist[idt+1]+'.h5')

    """
    Start downloading.
    """
    for ista in stalist:
        print('Downloading '+net+"."+ista+" ...")
        t0=time.time()
        """
        3a. Request data.
        """
        tr1,tr2,trZ,trP = obs.getdata(net,ista,sdatetime,edatetime,source=source,samp_freq=samp_freq,
                                      plot=False,rmresp=rmresp,pre_filt=pre_filt)
        sta_inv=client.get_stations(network=net,station=ista,
                                    starttime=sdatetime,endtime=edatetime,
                                    location='*',level='response')
        ta=time.time() - t0
        print('  downloaded '+net+"."+ista+" in "+str(ta)+" seconds.")
        """
        3b. Save to ASDF file.
        """
        tags=[]
        for itr,tr in enumerate([tr1,tr2,trZ,trP],1):
            if len(tr.stats.location) == 0:
                tlocation='00'
            else:
                tlocation=tr.stats.location

            tags.append(tr.stats.channel.lower()+'_'+tlocation.lower())

        print('  saving to '+fname)
        utils.save2asdf(fname,Stream(traces=[tr1,tr2,trZ,trP]),tags,sta_inv=sta_inv)



import numpy as np
from seisgo import utils
def fd1d_dx4dt4(x,dt,tmax,vmodel,rho,xsrc,xrcv,stf_freq=1,stf_shift=None,stf_type='ricker',t_interval=1):
    """
    Modified from Florian Wittkamp

    Finite-Difference acoustic seismic wave simulation
    Discretization of the first-order acoustic wave equation

    Temporal second-order accuracy $O(\Delta T^4)$

    Spatial fourth-order accuracy  $O(\Delta X^4)$

    Temporal discretization is based on the Adams-Basforth method is available in:

    Bohlen, T., & Wittkamp, F. (2016).
    Three-dimensional viscoelastic time-domain finite-difference seismic modelling using the staggered Adams-Bashforth time integrator.

    Geophysical Journal International, 204(3), 1781-1788.

    =====PARAMETERS=====
    x: spatial vector
    dt: time step for simulation.
    tmax: maximum time for simulation.
    vmodel: velocity model for each spatial grid.
    rho: density model for each spatial grid.
    xsrc: src grid index.
    xrcv: receiver grid index.
    stf_freq: source time function frequency parameter. Gaussian: width; Ricker: central frequency.
    stf_shift: source time function shift. default is 3*stf_freq.
    stf_type: "ricker" or "gaussian". Currently only support ricker.
    t_inverval: time interval of the output waveform. default: 1.

    ======RETURNS======
    tout,seisout: time and seismogram output.
    """
    c2=0.5  # CFL-Number. Stability condition.
    cmax=max(vmodel.flatten())
    dt_max=np.max(np.diff(x))/(cmax)*c2 #maximum time interval/step.
    if stf_shift is None:
        stf_shift = 3*stf_width
    nx=len(x)
    dx=np.abs(x[1]-x[0])

    if dt > dt_max:
        raise ValueError('dt %f is larger than allowable %f. '%(dt,dt_max))
    t=np.arange(0,tmax+stf_shift+0.5*dt,dt)     # Time vector
    nt=len(t)
    #wavelet information.
    # q0=1
    wavelet = np.zeros((len(t)))

    if stf_type.lower() == 'ricker':
        wlet0=utils.ricker(dt,stf_freq,stf_shift)[1]
        # tau=np.pi*stf_width*(t-1.5/stf_width)
        # wavelet=q0*(1.0-2.0*tau**2.0)*np.exp(-tau**2)
    elif stf_type.lower() == 'gaussian' or stf_type.lower() == 'gauss':
        wlet0=utils.gaussian(dt,stf_freq,stf_shift)[1]
    else:
        raise ValueError(stf_type+" not recoganized.")
    #
    wavelet[:len(wlet0)]=wlet0

    # Plotting source signal
#     plt.figure(figsize=(10,3))
#     plt.plot(t,wavelet)
#     plt.title('Source signal Ricker-Wavelet')
#     plt.ylabel('Amplitude')
#     plt.xlabel('Time in s')
#     plt.xlim(0,10)
#     plt.draw()

    # Init wavefields
    vx=np.zeros(nx)
    p=np.zeros(nx)
    vx_x=np.zeros(nx)
    p_x=np.zeros(nx)
    vx_x2=np.zeros(nx)
    p_x2=np.zeros(nx)
    vx_x3=np.zeros(nx)
    p_x3=np.zeros(nx)
    vx_x4=np.zeros(nx)
    p_x4=np.zeros(nx)

    # Calculate first Lame-Paramter
    l=rho * vmodel * vmodel

    ## Time stepping

    # Init Seismograms
    seisout=np.zeros((nt)); # Three seismograms

    # Calculation of some coefficients
    i_dx=1.0/(dx)
    kx=np.arange(0,nx-4)

    print("Starting time stepping...")
    ## Time stepping
    for n in range(2,nt):

        # Inject source wavelet
        p[xsrc]=p[xsrc]+wavelet[n]

        # Calculating spatial derivative
        p_x[kx]=i_dx*9.0/8.0*(p[kx+1]-p[kx])-i_dx*1.0/24.0*(p[kx+2]-p[kx-1])

        # Update velocity
        vx[kx]=vx[kx]-dt/rho[kx]*(13.0/12.0*p_x[kx]-5.0/24.0*p_x2[kx]+1.0/6.0*p_x3[kx]-1.0/24.0*p_x4[kx])

        # Save old spatial derivations for Adam-Bashforth method
        np.copyto(p_x4,p_x3)
        np.copyto(p_x3,p_x2)
        np.copyto(p_x2,p_x)

        # Calculating spatial derivative
        vx_x[kx]= i_dx*9.0/8.0*(vx[kx]-vx[kx-1])-i_dx*1.0/24.0*(vx[kx+1]-vx[kx-2])

        # Update pressure
        p[kx]=p[kx]-l[kx]*dt*(13.0/12.0*vx_x[kx]-5.0/24.0*vx_x2[kx]+1.0/6.0*vx_x3[kx]-1.0/24.0*vx_x4[kx])

        # Save old spatial derivations for Adam-Bashforth method
        np.copyto(vx_x4,vx_x3)
        np.copyto(vx_x3,vx_x2)
        np.copyto(vx_x2,vx_x)

        # Save seismograms
        seisout[n]=p[xrcv]

    print("Finished time stepping!")
    #shift to account for the stf_shift.

    if t_interval >1: #downsample data in time.
        tout=np.arange(0,tmax+0.5*t_interval*dt,t_interval*dt)
        seisout=np.interp(tout,t-stf_shift,seisout)
    else:
        tout=t[int(stf_shift/dt):]
        seisout=seisout[int(stf_shift/dt):]

    return tout,seisout

###
def build_vmodel(zmax,dz,nlayer,vmin,vmax,rhomin,rhomax,zmin=0,layer_dv=None):
    """
    Build layered velocity model with linearly increasing velocity, with the option of specifying anomalous layers.
    =========
    zmax: maximum depth of the model.
    dz: number of model grids, not the velocity layers. this is to create a fine grid layered model.
        with multiple grids within each layer.
    nlayer: number of velocity layers.
    vmin, vmax: velocity range.
    rhomin,rhomax: density range.
    zmin=0: minimum depth. default is 0.
    layer_dv: velocity perturbation for each layer.
    """
    layerv=np.linspace(vmin,vmax,nlayer)
    if layer_dv is None:
        layer_dv = np.zeros((nlayer))
    layerv = np.multiply(layerv,1+layer_dv)
    layerrho=np.linspace(rhomin,rhomax,nlayer)
    
    z=np.arange(zmin,zmax+-.5*dz,dz)
    
    zlayer=np.linspace(zmin,zmax,nlayer)
    v=np.zeros((len(z)))
    rho=np.zeros((len(z)))
    for i in range(len(z)):
        zidx_all=np.where((zlayer<=z[i]))[0]
        zidx=np.argmax(zlayer[zidx_all])
        
        v[i]=layerv[zidx]
        rho[i]=layerrho[zidx]
    #
    return z,v,rho



##
import pickle,os,math
import matplotlib.pyplot as plt
import plotly.express as px
import pandas as pd
import numpy as np
from seisgo import utils
from tslearn.utils import to_time_series, to_time_series_dataset
from tslearn.clustering import TimeSeriesKMeans
from minisom import MiniSom
from kneed import KneeLocator
######
def vpcluster_evaluate_kmean(ts,nrange,smooth=False,smooth_n=3,plot=True,njob=1,
                        metric='euclidean',max_iter_barycenter=100, random_state=0):
    """

    """
    distortion=[]
    for n in nrange:
        distortion.append(TimeSeriesKMeans(n_clusters=int(n), n_jobs=njob,\
                metric=metric, max_iter_barycenter=max_iter_barycenter, \
                random_state=random_state,verbose=False).fit(ts).inertia_)

    if smooth:
        ys=utils.box_smooth(distortion,smooth_n)
        ys[0]=distortion[0]
        ys[-1]=distortion[-1]
    else:
        ys = distortion


    nbest=list(KneeLocator(nrange, ys, S=1, curve="convex", direction="decreasing").all_knees)[0]
    if plot:
        plt.figure(figsize=(8,4),facecolor='w')
        plt.plot(nrange,distortion,'o',label='data')
        if smooth:
            plt.plot(nrange,ys,'r-',label='smoothed')
        else:
            plt.plot(nrange,ys,'r-')
        plt.vlines(nbest,np.min(ys),np.max(ys),label='knee:'+str(nbest))
        plt.xlabel('number of clusters',fontsize=12)
        plt.ylabel('sum of distance to center',fontsize=12)
        plt.xticks(nrange,fontsize=12)
        plt.yticks(fontsize=12)
        plt.legend()
        plt.show()

    return nbest,ys
def vpcluster_kmean(lat, lon, dep,vmodel,ncluster=None,nrange=None,spacing=1,njob=1,zrange=None,dz=None,
                         verbose=False,plot=True,savefig=True,figbase='kmean',
                      metric='euclidean',max_iter_barycenter=100, random_state=0,save=True,
                      source='vmodel',tag='v',figsize=None,evaluate_smooth=False,evaluate_plot=True):
    """
    zrange: target depth range for clustering. Default None, will use full range.
    dz: depth grid interval. If given, will interpolate the depth profiles.
    """
    if zrange is None:
        zrange=[np.min(dep),np.max(dep)]

    didx=np.where((dep>= zrange[0]) & (dep<= zrange[1]))[0]
    v=vmodel[didx]
    depth=dep[didx]

    if dz is not None: #do interpolation at depth direction.
        print("interpolation of depth at the interval of: %6.2f km"%(dz))
        zvec=np.arange(np.min(depth),np.max(depth)+0.5*dz,dz)
    else:
        zvec = depth

    all_v = []
    lat_subidx=[int(x) for x in np.arange(0,len(lat),spacing)]
    lon_subidx=[int(x) for x in np.arange(0,len(lon),spacing)]
    lat0=[]
    lon0=[]
    count=0
    for i in lat_subidx:
        for j in lon_subidx:
            v0=np.ndarray((v.shape[0]))
            for pp in range(v.shape[0]):
                v0[pp]=v[pp,i,j]
            if not np.isnan(v0).any() :
                if dz is not None:
                    vtemp=np.interp(zvec,depth,v0)
                    all_v.append(vtemp)
                else:
                    all_v.append(v0)
                lat0.append(lat[i])
                lon0.append(lon[j])
                count += 1

    ts = to_time_series_dataset(all_v)

    # determine the best number of clusters if ncluster is None.
    ss=[]
    if ncluster is None:
        print('ncluster is None. Determine the best. This may take a few minutes.')
        if nrange is None:
            nrange=np.arange(2,21,1)
        ncluster,ss = vpcluster_evaluate_kmean(ts,nrange,smooth=evaluate_smooth,smooth_n=3,plot=evaluate_plot,njob=njob,
                                metric=metric,max_iter_barycenter=max_iter_barycenter,
                                random_state=random_state)
    km = TimeSeriesKMeans(n_clusters=ncluster, n_jobs=njob,metric=metric, verbose=verbose,
                          max_iter_barycenter=max_iter_barycenter, random_state=random_state)
    y_pred = km.fit_predict(ts)


    rows = []
    for c in range(count):
        cluster = km.labels_[c]
        rows.append([lat0[c], lon0[c], cluster+1])

    df = pd.DataFrame(rows, columns=['lat', 'lon', 'cluster'])
    cdata=[]
    for yi in range(ncluster):
        cdata.append(ts[y_pred == yi].T)

    outdict=dict()
    outdict['method']="k-means"
    outdict['source']=source
    outdict['tag']=tag
    outdict['depth']=zvec
    outdict['model']=km
    outdict['pred']=cdata
    outdict['para']={'n_clusters':ncluster,'n_jobs':njob,'metric':metric,
                        'max_iter_barycenter':max_iter_barycenter,
                        'random_state':random_state,'nrange':nrange,'sum_square':ss}
    outdict['cluster_map']=df

    outfile=figbase+"_clusters_k"+str(ncluster)+"_results.pk"
    if save:
        with open(outfile,'wb') as f:
            pickle.dump(outdict,f)

    if plot:
        ###########
        #### plotting clustered data/time series.
        ###########
        if figsize is None:
            if ncluster<4:
                plt.figure(figsize=(13, 4),facecolor='w')
            else:
                plt.figure(figsize=(13, 9),facecolor='w')
        else:
            plt.figure(figsize=figsize,facecolor='w')
        for yi in range(ncluster):
            if ncluster<4:
                plt.subplot(1, ncluster, yi + 1)
            elif ncluster<9:
                plt.subplot(int(np.ceil(ncluster/2)), 2, yi + 1)
            elif ncluster < 16:
                plt.subplot(int(np.ceil(ncluster/3)), 3, yi + 1)
            elif ncluster < 21:
                plt.subplot(int(np.ceil(ncluster/4)), 4, yi + 1)
            elif ncluster < 26:
                plt.subplot(int(np.ceil(ncluster/5)), 5, yi + 1)
            else:
                plt.subplot(int(np.ceil(ncluster/6)), 6, yi + 1)
            for xx in cdata[yi]:
                plt.plot(zvec,xx, "k-", alpha=.2)
            plt.plot(zvec,km.cluster_centers_[yi].ravel(), "r-")
            plt.text(0.65, 0.15, 'Cluster %d' % (yi + 1),
                     transform=plt.gca().transAxes)
            plt.title(f"Cluster {yi+1}")
            plt.xlabel('depth (km)')
            plt.ylabel('Vs (km/s)')
        plt.tight_layout()

        if savefig:
            plt.savefig(figbase+"_clusters_k"+str(ncluster)+".png",format="png")
            plt.close()
        else:
            plt.show()

        #####################
        ######## plot map view of clusters.
        ####################
        # Create map using plotly
        fig = px.scatter_mapbox(
            df,lat="lat",lon='lon',color='cluster',size_max=13,zoom=3,width=900,height=800,
        )

        fig.update_layout(
            mapbox_style="white-bg",
            mapbox_layers=[
                {
                    "below": 'traces',
                    "sourcetype": "raster",
                    "source": ["https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryOnly/MapServer/tile/{z}/{y}/{x}"]
                }
            ]
        )
        if savefig:
            fig.write_image(figbase+"_clustermap_k"+str(ncluster)+".png",format="png")
        else:
            fig.show()
    #
    if not save:
        return outdict

#
def vpcluster_som(lat, lon, depth,v,grid_size=None,spacing=1,niteration=50000,sigma=0.3,
                     rate=0.1,verbose=False,plot=True,savefig=True,figbase='som',
                      save=True,source='vmodel',tag='v',figsize=None):
    all_v = []
    lat_subidx=[int(x) for x in np.arange(0,len(lat),spacing)]
    lon_subidx=[int(x) for x in np.arange(0,len(lon),spacing)]
    lat0=[]
    lon0=[]
    count=0
    for i in lat_subidx:
        for j in lon_subidx:
            v0=np.ndarray((v.shape[0]))
            for pp in range(v.shape[0]):
                v0[pp]=v[pp,i,j]
            if not np.isnan(v0).any() :
                all_v.append(v0)
                lat0.append(lat[i])
                lon0.append(lon[j])
                count += 1

    if grid_size is None:
        som_x = som_y = math.ceil(math.sqrt(math.sqrt(len(all_v))))
    else:
        som_x=grid_size[0]
        som_y=grid_size[1]
    som = MiniSom(som_x, som_y,len(all_v[0]), sigma=sigma, learning_rate = rate)
    som.random_weights_init(all_v)
    som.train(all_v, niteration)

    win_map = som.win_map(all_v)
    # Returns the mapping of the winner nodes and inputs

    rows = []
    for c in range(count):
        wm=som.win_map([all_v[c]])
        x,y=list(wm.keys())[0]
        cluster=x*som_y + y + 1
        rows.append([lat0[c], lon0[c], cluster])
    df = pd.DataFrame(rows, columns=['lat', 'lon', 'cluster'])

    #clustered data
    cdata=[]

    for x in range(som_x):
        for y in range(som_y):
            cluster = (x,y)
            if cluster in win_map.keys():
                cdata.append(win_map[cluster])

    outdict=dict()
    outdict['method']="som"
    outdict['source']=source
    outdict['tag']=tag
    outdict['depth']=depth
    outdict['model']=som
    outdict['pred']=cdata
    outdict['para']={'nx':som_x,'ny':som_y,'sigma':sigma,
                        'niteration':niteration, 'learning_rate':rate}
    outdict['cluster_map']=df

    outfile=figbase+"_clusters_k"+str(som_x)+"x"+str(som_y)+"_results.pk"
    if save:
        with open(outfile,'wb') as f:
            pickle.dump(outdict,f)

    ###########
    #######
    if plot:
        ncluster=len(cdata)
        if figsize is None:
            if ncluster<4:
                plt.figure(figsize=(13, 4),facecolor='w')
            else:
                plt.figure(figsize=(13, 9),facecolor='w')
        else:
            plt.figure(figsize=figsize,facecolor='w')
        for i in range(len(cdata)):
            if ncluster<4:
                plt.subplot(1, ncluster, i + 1)
            elif ncluster<9:
                plt.subplot(int(np.ceil(ncluster/2)), 2, i + 1)
            elif ncluster < 16:
                plt.subplot(int(np.ceil(ncluster/3)), 3, i + 1)
            elif ncluster < 21:
                plt.subplot(int(np.ceil(ncluster/4)), 4, i + 1)
            elif ncluster < 26:
                plt.subplot(int(np.ceil(ncluster/5)), 5, i + 1)
            else:
                plt.subplot(int(np.ceil(ncluster/6)), 6, i + 1)

            for series in cdata[i]:
                plt.plot(depth,series,c="gray",alpha=0.3)
            plt.plot(depth,np.average(np.vstack(cdata[i]),axis=0),c="red")
            plt.title(f"Cluster {i+1}")
            plt.xlabel('depth (km)')
            plt.ylabel('Vs (km/s)')
        plt.tight_layout()
        if savefig:
            plt.savefig(figbase+"_clusters_k"+str(som_x)+"x"+str(som_y)+".png",format="png")
            plt.close()
        else:
            plt.show()

        #####################
        ######## plot map view of clusters.
        ####################
        # Create map using plotly
        fig = px.scatter_mapbox(
            df,lat="lat",lon='lon',color='cluster',size_max=13,zoom=3,width=900,height=800,
        )

        fig.update_layout(
            mapbox_style="white-bg",
            mapbox_layers=[
                {
                    "below": 'traces',
                    "sourcetype": "raster",
                    "source": ["https://basemap.nationalmap.gov/arcgis/rest/services/USGSImageryOnly/MapServer/tile/{z}/{y}/{x}"]
                }
            ]
        )
        if savefig:
            fig.write_image(figbase+"_clustermap_k"+str(som_x)+"x"+str(som_y)+".png",format="png")
        else:
            fig.show()
    if not save:
        return outdict



import os,glob,copy,obspy,scipy,time
import numpy as np
from seisgo.utils import rms
import matplotlib.pyplot as plt
from scipy.signal import hilbert
from scipy.fftpack import fft,ifft,next_fast_len
from stockwell import st
from tslearn.utils import to_time_series, to_time_series_dataset
from tslearn.clustering import TimeSeriesKMeans
"""
Stacking functions.
"""
def stack(d,method,par=None):
    """
    this is a wrapper for calling individual stacking functions.
    d: data. 2-d array
    method: stacking method, one of "linear","pws","robust","acf","nroot","selective",
            "cluster"
    par: dictionary containing all parameters for each stacking method. defaults will
        be used if not specified.

    RETURNS:
    ds: stacked data, which may be a list depending on the method.
    """
    #remove NaN traces.
    newdata= [] 
    if d.ndim >1:
        for d2 in d: 
            if not np.isnan(d2).any(): 
                newdata.append(d2)
        newdata = np.array(newdata)
    else:
        newdata=d.copy()

    method_list=["linear","pws","robust","acf","nroot","selective",
            "cluster","tfpws","tfpws-dost"]
    if method not in method_list:
        raise ValueError("$s not recoganized. use one of $s"%(method,str(method_list)))
    par0={"axis":0,"p":2,"g":1,"cc_min":0.0,"epsilon":1E-5,"maxstep":10,
            "win":None,"stat":False,"h":0.75,'plot':False,'normalize':True,'ref':None}  #stat: if true, will return statistics.
    if par is None:
        par=par0
    else:
        par={**par0,**par} #use par values if specified. otherwise, use defaults.

    if method.lower() == 'linear':
        ds = np.mean(newdata,axis=par["axis"])
    elif method.lower() == 'pws':
        ds = pws(newdata,p=par['p'])
    elif method.lower() == 'tfpws':
        ds = tfpws(newdata,p=par['p'])
    elif method.lower() == 'tfpws-dost':
        ds = tfpws_dost(newdata,p=par['p'])
    elif method.lower() == 'robust':
        ds = robust(newdata,epsilon=par['epsilon'],maxstep=par['maxstep'],win=par["win"],
                stat=par['stat'],ref=par['ref'])
    elif method.lower() == 'acf':
        ds = adaptive_filter(newdata,g=par['g'])
    elif method.lower() == 'nroot':
        ds = nroot(newdata,p=par['p'])
    elif method.lower() == 'selective':
        ds = selective(newdata,cc_min=par['cc_min'],epsilon=par['epsilon'],maxstep=par['maxstep'],
                stat=par['stat'],ref=par['ref'],win=par["win"])
    elif method.lower() == 'cluster':
        ds = clusterstack(newdata,h=par['h'],axis=par['axis'],win=par["win"],
        normalize=par['normalize'],plot=par['plot'])
    #
    return ds

def seisstack(d,method,par=None):
    """
    This is the same as stack(), to be compatible with old usage.
    """
    return stack(d,method=method,par=par)

def robust(d,epsilon=1E-5,maxstep=10,win=None,stat=False,ref=None):
    """
    this is a robust stacking algorithm described in Pavlis and Vernon 2010. Generalized
    by Xiaotao Yang.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    epsilon: residual threhold to quit the iteration (a small number). Default 1E-5
    maxstep: maximum iterations. default 10.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    ref: reference stack, with the same length as individual data. Default: None. Use median().
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation
    w: weight (return only when stat = True.)
    nstep: number of iterations to produce the final stack (return only when stat = True.)
    Written by Marine Denolle
    Modified by Xiaotao Yang
    """
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    res  = 9E9  # residuals
    w = np.ones(d.shape[0])
    small_number=1E-15
    max_abs_value=np.max(np.abs(d)) 
    dcopy=d/max_abs_value #to avoid the dot product and L2 normm getting too large.
    nstep=0
    if N >=2:
        if ref is None:
            newstack = np.median(dcopy,axis=0)
        else:
            newstack = ref
        if win is None:
            win=[0,-1]
        while res > epsilon and nstep <=maxstep:
            stackt = newstack
            for i in range(dcopy.shape[0]):
                dtemp=dcopy[i,win[0]:win[1]]
                crap = np.multiply(stackt[win[0]:win[1]],dtemp.T)
                crap_dot = np.sum(crap)
                di_norm = np.linalg.norm(dtemp)
                ri_norm = np.linalg.norm(dtemp -  crap_dot*stackt[win[0]:win[1]])
                if ri_norm < small_number:
                    w[i]=0
                else:
                    w[i]  = np.abs(crap_dot) /di_norm/ri_norm
            w =w /np.sum(w)
            newstack =np.sum( (w*dcopy.T).T,axis=0)#/len(cc_array[:,1])
            res = np.linalg.norm(newstack-stackt,ord=1)/np.linalg.norm(newstack)/len(dcopy[:,1])
            nstep +=1
    else:
        newstack=dcopy[0].copy()
    newstack *= max_abs_value #scale the stack back.
    if stat:
        return newstack, w, nstep
    else:
        return newstack

def adaptive_filter(d,g=1):
    '''
    the adaptive covariance filter to enhance coherent signals. Fellows the method of
    Nakata et al., 2015 (Appendix B)

    the filtered signal [x1] is given by x1 = ifft(P*x1(w)) where x1 is the ffted spectra
    and P is the filter. P is constructed by using the temporal covariance matrix.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D traces of daily/hourly cross-correlation functions
    g: a positive number to adjust the filter harshness [default is 1]
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation function
    '''
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N>=2:
        Nfft = next_fast_len(M)

        # fft the 2D array
        spec = fft(d,axis=1,n=Nfft)[:,:M]

        # make cross-spectrm matrix
        cspec = np.zeros(shape=(N*N,M),dtype=np.complex64)
        for ii in range(N):
            for jj in range(N):
                kk = ii*N+jj
                cspec[kk] = spec[ii]*np.conjugate(spec[jj])

        S1 = np.zeros(M,dtype=np.complex64)
        S2 = np.zeros(M,dtype=np.complex64)
        # construct the filter P
        for ii in range(N):
            mm = ii*N+ii
            S2 += cspec[mm]
            for jj in range(N):
                kk = ii*N+jj
                S1 += cspec[kk]

        p = np.power((S1-S2)/(S2*(N-1)),g)

        # make ifft
        narr = np.real(ifft(np.multiply(p,spec),Nfft,axis=1)[:,:M])
        newstack=np.mean(narr,axis=0)
    else:
        newstack=d[0].copy()
    #
    return newstack

def pws(d,p=2):
    '''
    Performs phase-weighted stack on array of time series. Modified on the noise function by Tim Climents.
    Follows methods of Schimmel and Paulssen, 1997.
    If s(t) is time series data (seismogram, or cross-correlation),
    S(t) = s(t) + i*H(s(t)), where H(s(t)) is Hilbert transform of s(t)
    S(t) = s(t) + i*H(s(t)) = A(t)*exp(i*phi(t)), where
    A(t) is envelope of s(t) and phi(t) is phase of s(t)
    Phase-weighted stack, g(t), is then:
    g(t) = 1/N sum j = 1:N s_j(t) * | 1/N sum k = 1:N exp[i * phi_k(t)]|^v
    where N is number of traces used, v is sharpness of phase-weighted stack

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    p: exponent for phase stack (int). default is 2

    RETURNS:
    ---------------------
    newstack: Phase weighted stack of time series data (numpy.ndarray)
    '''

    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >=2:
        analytic = hilbert(d,axis=1, N=next_fast_len(M))[:,:M]
        phase = np.angle(analytic)
        phase_stack = np.mean(np.exp(1j*phase),axis=0)
        phase_stack = np.abs(phase_stack)**(p)

        weighted = np.multiply(d,phase_stack)

        newstack=np.mean(weighted,axis=0)
    else:
        newstack=d[0].copy()
    return newstack

def nroot(d,p=2):
    '''
    this is nth-root stacking algorithm translated based on the matlab function
    from https://github.com/xtyangpsp/SeisStack (by Xiaotao Yang; follows the
    reference of Millet, F et al., 2019 JGR)

    Parameters:
    ------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    p: np.int, nth root for the stacking. Default is 2.

    Returns:
    ------------
    newstack: np.ndarray, final stacked waveforms

    Written by Chengxin Jiang @ANU (May2020)
    '''
    if d.ndim == 1:
        print('2D matrix is needed for nroot_stack')
        return d
    N,M = d.shape
    if N >=2:
        dout = np.zeros(M,dtype=np.float32)

        # construct y
        for ii in range(N):
            dat = d[ii,:]
            dout += np.sign(dat)*np.abs(dat)**(1/p)
        dout /= N

        # the final stacked waveform
        newstack = dout*np.abs(dout)**(p-1)
    else:
        newstack=d[0].copy()

    return newstack


def selective(d,cc_min,epsilon=1E-5,maxstep=10,win=None,stat=False,ref=None):
    '''
    this is a selective stacking algorithm developed by Jared Bryan/Kurama Okubo.

    PARAMETERS:
    ----------------------
    d: numpy.ndarray contains the 2D cross correlation matrix
    epsilon: residual threhold to quit the iteration
    cc_min: numpy.float, threshold of correlation coefficient to be selected
    epsilon: residual threhold to quit the iteration (a small number). Default 1E-5
    maxstep: maximum iterations. default 10.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    ref: reference stack, with the same length as individual data. Default: None. Use mean().
    RETURNS:
    ----------------------
    newstack: numpy vector contains the stacked cross correlation
    nstep: np.int, total number of iterations for the stacking

    Originally ritten by Marine Denolle
    Modified by Chengxin Jiang @Harvard (Oct2020)
    '''
    if d.ndim == 1:
        print('2D matrix is needed for selective stacking')
        return d
    N,M = d.shape
    if N>=2:
        res  = 9E9  # residuals
        cof  = np.zeros(N,dtype=np.float32)
        if ref is None:
            newstack = np.mean(d,axis=0)
        else:
            newstack = ref

        nstep = 0
        if win is None:
            win=[0,-1]
        # start iteration
        while res>epsilon and nstep<=maxstep:
            for ii in range(N):
                cof[ii] = np.corrcoef(newstack[win[0]:win[1]], d[ii,win[0]:win[1]])[0, 1]

            # find good waveforms
            indx = np.where(cof>=cc_min)[0]
            nstep +=1
            if not len(indx):
                newstack=np.ndarray((d.shape[1],))
                newstack.fill(np.nan)
                print('cannot find good waveforms inside selective stacking')
                break
            else:
                oldstack = newstack
                newstack = np.mean(d[indx],axis=0)
                res = np.linalg.norm(newstack-oldstack)/(np.linalg.norm(newstack)*M)
    else:
        newstack=d[0].copy()
    if stat:
        return newstack, nstep
    else:
        return newstack
#
def clusterstack(d,h=0.75,win=None,axis=0,normalize=True,plot=False):
    '''
    Performs stack after clustering. The data will be clustered into two groups.
    If the two centers of the clusters are similar (defined by corrcoef >= "t"), the original
    traces associated with both clusters will be used to produce the final linear stack, weighted by
    normalized SNR (phase clarity) of each cluster. Otherwise, the one with larger phase clarity
    (defined as max(abs(amplitudes))/rms(abs(amplitudes))) will be used to get the final stack.

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    h: corrcoeff threshold to decide which group/cluster to use. Default 0.75.
    win: [start_index,end_index] used to compute the weight, instead of the entire trace. Default None.
            When None, use the entire trace.
    axis: which axis to stack. default 0.
    normalize: Normalize the traces before clustering. This will only influence the cluster.
            The final stack will be produced using the original data.
    plot: plot clustering results. default False.

    RETURNS:
    ---------------------
    newstack: final stack.
    '''
    ncluster=2 #DO NOT change this value.
    min_trace=2 #minimum of two traces.
    metric="euclidean" #matric to compute the distance in kmeans clustering.
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >= min_trace:
        dataN=d.copy()
        if normalize:
            for i in range(N):
                dataN[i]=d[i]/np.max(np.abs(d[i]),axis=0)

        ts = to_time_series_dataset(dataN)

        km = TimeSeriesKMeans(n_clusters=ncluster, n_jobs=1,metric=metric, verbose=False,
                              max_iter_barycenter=100, random_state=0)
        y_pred = km.fit_predict(ts)
        snr_all=[]
        centers_all=[]
        cidx=[]
        if win is None:
            win=[0,-1]
        for yi in range(ncluster):
            cidx.append(np.where((y_pred==yi))[0])
            center=km.cluster_centers_[yi].ravel()#np.squeeze(np.mean(ts[y_pred == yi].T,axis=2))
            centers_all.append(center)
            snr=np.max(np.abs(center[win[0]:win[1]]))/rms(np.abs(center))
            snr_all.append(snr)

        #
        if plot:
            plt.figure(figsize=(12,4))
            for yi in range(ncluster):
                plt.subplot(1,ncluster,yi+1)
                plt.plot(np.squeeze(ts[cidx[yi]].T),'k-',alpha=0.3)
                plt.plot(centers_all[yi],'r-')
                plt.title('Cluster %d: %d'%(yi+1,len(cidx[yi])))
            plt.show()
        cc=np.corrcoef(centers_all[0],centers_all[1])[0,1]
        if cc>= h: #use all data
            snr_normalize=snr_all/np.sum(snr_all)
            newstack=np.zeros((M))
            for yi in range(ncluster):
                newstack += snr_normalize[yi]*np.mean(d[cidx[yi]],axis=0)
        else:
            goodidx=np.argmax(snr_all)
            newstack=np.mean(d[cidx[goodidx]],axis=0)
        del dataN,ts,y_pred
    else:
        newstack=d[0].copy()
    #
    return newstack

def tfpws(d,p=2,axis=0):
    '''
    Performs time-frequency domain phase-weighted stack on array of time series.

    $C_{ps} = |(\sum{S*e^{i2\pi}/|S|})/M|^p$, where $C_{ps}$ is the phase weight. Then
    $S_{pws} = C_{ps}*S_{ls}$, where $S_{ls}$ is the S transform of the linea stack
    of the whole data.

    Reference:
    Schimmel, M., Stutzmann, E., & Gallart, J. (2011). Using instantaneous phase
    coherence for signal extraction from ambient noise data at a local to a
    global scale. Geophysical Journal International, 184(1), 494–506.
    https://doi.org/10.1111/j.1365-246X.2010.04861.x

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    p: exponent for phase stack (int). default is 2
    axis: axis to stack, default is 0.

    RETURNS:
    ---------------------
    newstack: Phase weighted stack of time series data (numpy.ndarray)
    '''
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >=2:
        lstack=np.mean(d,axis=axis)
        #get the ST of the linear stack first
        stock_ls=st.st(power2pad(lstack))

        #run a ST to get the dimension of ST result
        stock_temp=st.st(power2pad(d[0]))
        phase_stack=np.zeros((stock_temp.shape[0],stock_temp.shape[1]),dtype='complex128')
        for i in range(N):
            if i>0: #zero index has been computed.
                stock_temp=st.st(power2pad(d[i]))
            phase_stack += np.multiply(stock_temp,np.angle(stock_temp))/np.abs(stock_temp)
        #
        phase_stack = np.abs(phase_stack/N)**p

        pwstock=np.multiply(phase_stack,stock_ls)
        recdostIn=np.real(st.ist(pwstock))
        newstack=recdostIn[:M] # trim padding
    else:
        newstack=d[0].copy()
    #
    return newstack

def tfpws_dost(d,p=2,axis=0):
    '''
    Performs time-frequency domain phase-weighted stack on array of time series using DOST (Discrete
    orthogonal stockwell transform).
    $C_{ps} = |(\sum{S*e^{i2\pi}/|S|})/M|^p$, where $C_{ps}$ is the phase weight. Then
	$S_{pws} = C_{ps}*S_{ls}$, where $S_{ls}$ is the Discrete Orthonormal S transform
	of the linear stack of the whole data.

    DOST stacking was implemented by Jared Bryan.

    Reference for tf-PWS:
    Schimmel, M., Stutzmann, E., & Gallart, J. (2011). Using instantaneous phase
    coherence for signal extraction from ambient noise data at a local to a
    global scale. Geophysical Journal International, 184(1), 494–506.
    https://doi.org/10.1111/j.1365-246X.2010.04861.x

    Reference for DOST:
    U. Battisti, L. Riba, "Window-dependent bases for efficient representations of the
	Stockwell transform", Applied and Computational Harmonic Analysis, 23 February 2015,
	http://dx.doi.org/10.1016/j.acha.2015.02.002.

    PARAMETERS:
    ---------------------
    d: N length array of time series data (numpy.ndarray)
    p: exponent for phase stack (int). default is 2
    axis: axis to stack, default is 0.

    RETURNS:
    ---------------------
    newstack: Phase weighted stack of time series data (numpy.ndarray)
    '''
    if d.ndim == 1:
        print('2D matrix is needed')
        return d
    N,M = d.shape
    if N >=2:
        lstack=np.mean(d,axis=axis)
        #get the dost of the linear stack first
        stock_ls_dost=DOST(lstack) # initialize dost object
        stock_ls=stock_ls_dost.dost(stock_ls_dost.data) # calculate the dost

    	# calculate dost for first trace to know its shape
        stock_dost=DOST(d[0])
        stock_temp=stock_dost.dost(stock_dost.data)
    	# initialize stack
        phase_stack=np.zeros(len(stock_temp),dtype='complex128')
    	# calculate the dost for each trace to be stacked
        for i in range(d.shape[0]):
        	if i>0: # zero index has been computed
        		stock_dost=DOST(d[i])
        		stock_temp=stock_dost.dost(stock_dost.data)
        	phase_stack+=np.multiply(stock_temp,np.angle(stock_temp))/np.abs(stock_temp)

        phase_stack = np.abs(phase_stack/N)**p

        pwstock=np.multiply(phase_stack,stock_ls)
        recdostIn = np.real(stock_dost.idost(pwstock))
        newstack = recdostIn[:M] # trim padding
    else:
        newstack=d[0].copy()
    #
    return newstack

#################################
####### stacking needed utilities.
################################
#
def power2pad(data):
	"""Zero pad data such that its length is a power of 2"""
	N=int(2**np.ceil(np.log2(len(data))))
	pad_end=np.zeros(int(N-len(data)))

	return np.concatenate((data,pad_end))
class DOST:
	def __init__(self, data):
		# make sure data length is a power of 2
		if np.ceil(np.log2(len(data)))==np.floor(np.log2(len(data))):
			# length of data already a power of 2
			self.data=data
		else:
			# pad data to nearest power of 2
			self.data=self.pad(data)

	def pad(self,data):
		"""Zero pad data such that its length is a power of 2"""
		N=int(2**np.ceil(np.log2(len(data))))
		pad_end=np.zeros(int(N-len(data)))
		data=np.concatenate((data,pad_end))

		return data

	def fourier(self, d):
		"""Normalize and center fft"""
		fftIn=(1/np.sqrt(len(d))) * np.fft.fftshift(np.fft.fft(np.fft.ifftshift(d)))
		return fftIn

	def ifourier(self,d):
		"""Normalize and center ifft"""
		ifftIn=np.sqrt(len(d)) * np.fft.fftshift(np.fft.ifft(np.fft.ifftshift(d)))
		return ifftIn

	def dostbw(self,D):
		"""Calculate size of the DOST bandwidths"""
		arr=[0]
		arr.extend(np.arange(np.log2(D)-2, -1e-9, -1))
		arr.extend([0])
		arr.extend(np.arange(0, np.log2(D)-2+1e-9))
		arr=2**np.array(arr)
		return arr

	def dost(self,d):
		"""Discrete Orthonormal Stockwell Transform"""
		d_dost=self.fourier(d)
		D=len(d)
		bw=self.dostbw(D)
		k=0
		for i in bw:
			i=int(i)
			if i==1:
				k=k+i
			else:
				d_dost[k:k+i] = self.ifourier(d_dost[k:k+i])
				k=k+i
		return d_dost

	def idost(self,d):
		"""Inverse Discrete Orthonormal Stockwell Transform"""
		d_idost=d
		D=len(d)
		bw=self.dostbw(D)
		k=0
		for i in bw:
			i=int(i)
			if i==1:
				k=k+i
			else:
				d_idost[k:k+i] = self.fourier(d_idost[k:k+i])
				k=k+i
		d_idost = self.ifourier(d_idost)
		return d_idost



import sys
import obspy
import os
import time
import numpy as np
import pandas as pd
from mpi4py import MPI
from seisgo.utils import split_datetimestr
from seisgo import downloaders,noise
#########################################################
################ PARAMETER SECTION ######################
#########################################################
tt0=time.time()

# paths and filenames
rootpath = "tempdata_craton" # roothpath for the project
direc  = os.path.join(rootpath,'Raw')                   # where to store the downloaded data
#if not os.path.isdir(direc): os.mkdir(direc)
down_list  = os.path.join(direc,'station.txt')
# CSV file for station location info

# download parameters
source='IRIS'                                 # client/data center. see https://docs.obspy.org/packages/obspy.clients.fdsn.html for a list
max_tries = 3                                                  #maximum number of tries when downloading, in case the server returns errors.
use_down_list = False                                                # download stations from a pre-compiled list or not
flag      = True                                               # print progress when running the script; recommend to use it at the begining
samp_freq = 5                                                  # targeted sampling rate at X samples per seconds
rmresp   = True
rmresp_out = 'VEL'
pressure_chan = [None]				#Added by Xiaotao Yang. This is needed when downloading some special channels, e.g., pressure data. VEL output for these channels.
respdir   = os.path.join(rootpath,'resp')                       # directory where resp files are located (required if rm_resp is neither 'no' nor 'inv')
freqmin   = 0.002                                                # pre filtering frequency bandwidth
freqmax   = 0.5*samp_freq
# note this cannot exceed Nquist freq

# targeted region/station information: only needed when use_down_list is False
lamin,lamax,lomin,lomax= 34,48,-98,-78.5                # regional box: min lat, max lat, min lon, max lon
chan_list = ["BHZ","HHZ"]
net_list  = ["XO","7A","CN","IM","IU","LD","LM","MU","N4","NM","OH","OK","PE","PN","SS","US","XI","Z9","ZL"]                   # network list
sta_list  = ["*"]                                               # station (using a station list is way either compared to specifying stations one by one)
start_date = "2013_09_28_0_0_0"                               # start date of download
end_date   = "2014_01_01_0_0_0"                               # end date of download
inc_hours  = 12                                                 # length of data for each request (in hour)
maxseischan = 1                                                  # the maximum number of seismic channels, excluding pressure channels for OBS stations.
ncomp      = maxseischan #len(chan_list)

# get rough estimate of memory needs to ensure it now below up in noise cross-correlations
cc_len    = 2*3600                                                # basic unit of data length for fft (s)
step      = 1*3600                                                 # overlapping between each cc_len (s)
MAX_MEM   = 5.0                                                 # maximum memory allowed per core in GB


##################################################
# we expect no parameters need to be changed below
# assemble parameters used for pre-processing waveforms in downloading
prepro_para = {'rmresp':rmresp,'rmresp_out':rmresp_out,'respdir':respdir,'freqmin':freqmin,'freqmax':freqmax,\
                'samp_freq':samp_freq}

downlist_kwargs = {"source":source, 'net_list':net_list, "sta_list":sta_list, "chan_list":chan_list, \
                    "starttime":start_date, "endtime":end_date, "maxseischan":maxseischan, "lamin":lamin, "lamax":lamax, \
                    "lomin":lomin, "lomax":lomax, "pressure_chan":pressure_chan, "fname":down_list}

########################################################
#################DOWNLOAD SECTION#######################
########################################################
#--------MPI---------
comm = MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

if rank==0:
    if flag:
        print('station.list selected [%s] for data from %s to %s with %sh interval'%(use_down_list,start_date,end_date,inc_hours))

    if not os.path.isdir(direc):os.makedirs(direc)
    if use_down_list:
        stalist=pd.read_csv(down_list)
    else:
        stalist=downloaders.get_sta_list(**downlist_kwargs) # saves station list to "down_list" file
                                              # here, file name is "station.txt"
    # rough estimation on memory needs (assume float32 dtype)
    memory_size=noise.cc_memory(inc_hours,samp_freq,len(stalist.station),ncomp,cc_len,step)
    if memory_size > MAX_MEM:
        raise ValueError('Require %5.3fG memory but only %5.3fG provided)! Reduce inc_hours to avoid this issue!' % (memory_size,MAX_MEM))

    # save parameters for future reference
    metadata = os.path.join(direc,'download_info.txt')
    fout = open(metadata,'w')
    fout.write(str({**prepro_para,**downlist_kwargs,'inc_hours':inc_hours,'ncomp':ncomp}));fout.close()

    all_chunk = split_datetimestr(start_date,end_date,inc_hours)
    if len(all_chunk)<1:
        raise ValueError('Abort! no data chunk between %s and %s' % (start_date,end_date))
    splits = len(all_chunk)-1
else:
    splits,all_chunk = [None for _ in range(2)]

# broadcast the variables
splits = comm.bcast(splits,root=0)
all_chunk  = comm.bcast(all_chunk,root=0)
extra = splits % size

# MPI: loop through each time chunk
for ick in range(rank,splits,size):
    s1= all_chunk[ick]
    s2=all_chunk[ick+1]

    download_kwargs = {"source":source,"rawdatadir": direc, "starttime": s1, "endtime": s2, \
              "stationinfo": down_list,**prepro_para}

    # Download for ick
    downloaders.download(**download_kwargs)

tt1=time.time()
print('downloading step takes %6.2f s' %(tt1-tt0))

comm.barrier()
#if rank == 0:
#    sys.exit()



import obspy,os,utm
import numpy as np
import pandas as pd
from seisgo import noise,utils
import matplotlib.pyplot as plt
from scipy.optimize import curve_fit
import pygmt as gmt
from matplotlib.colors import ListedColormap
import cartopy.crs as ccrs
import cartopy.feature as cfeature
import matplotlib.ticker as mticker
import matplotlib as mpl
mpl.rcParams['pdf.fonttype'] = 42
mpl.rcParams['ps.fonttype'] = 42
# mpl.rc('font',family='Helvetica')

"""
This module contains functions to compute anisotropy. 

As of now, it contains functions to compute the azimuthal anisotropy using BANX method (Beamforming Azimuthal Anisotropy 
using Noise Cross-correlation). The main worklow was converted from the original MATLAB codes by Jorge C. Castellanos-Martinez, 
provided by the author. See the related publication for more details.
Castellanos, J. C., Perry-Houts, J., Clayton, R. W., Kim, Y., Stanciu, A. C., Niday, B., & Humphreys, E. (2020). Seismic anisotropy 
reveals crustal flow driven by mantle vertical loading in the Pacific NW. Science Advances, 6(28), 1-10. https://doi.org/10.1126/sciadv.abb0476

Noted by Xiaotao Yang, March 11, 2025.
"""

def get_ArrayAttributes(lat,lon,get_utm=False):
    """
    Get the attributes of the station array. 

    ===PARAMETER==
    lat, lon: latitudes and longitudes.
    ===RETURN==
    radius: radius of the circle enclosing the stations/array. Unit: km
    center: center coordinates of the stations [lat, lon].

    ==PACKAGES==
    utm: to convert to and from the geographical coordinates.
    """

    #convert to lat lon to UTM to get the center.
    Ueast,Unorth,Zone,ZoneStr = utm.from_latlon(lat,lon)
    center = utm.to_latlon(np.mean(Ueast),np.mean(Unorth),Zone,ZoneStr)

    #compute distance of each station to the center.
    dist=[]
    for i in range(len(lat)):
        dist.append(obspy.geodetics.base.gps2dist_azimuth(center[0],center[1],
                                                    lat[i],
                                                    lon[i])[0]/1000)
    #
    radius = np.max(dist)

    #
    if get_utm: #return UTM outputs for all stations.
        return radius,center,[Ueast,Unorth,Zone,ZoneStr]
    else:
        return radius,center
    
def compute_anisotropy(x, a, b, c):
    """
    Compute anisotropic velocity. Only use the first three terms (the 2*theta terms).

    $v = a + b*cos(2*\theta) + c*sin(2*\theta) $
    theta: direction/azimuth. Independent variable: x.
    a: isotropic velocity.
    b and c: coefficients of the anisotropic directions.
    v: anisotropic velocity.
    ===PARAMETER===
    x: direction/azimuth in degrees. Independent variable.
    a: isotropic velocity.
    b: coefficient of the anisotropic direction.
    c: coefficient of the anisotropic direction.
    ===RETURN===
    v: anisotropic velocity.
    ===REFERENCE===
    M. L. Smith, F. A. Dahlen, The azimuthal dependence of Love and Rayleigh wave
        propagation in a slightly anisotropic medium. J. Geophys. Res. 78, 3321–3333 (1973).
    """
    return a + b*np.cos(2*np.deg2rad(x)) + c*np.sin(2*np.deg2rad(x))

#################### BANX method ####################
def do_BANX(stationdict_all, reference_site, period_band, reference_velocity, datadir,outdir_root='.',sampling_rate=None,
            min_stations=10, min_snr=5, min_radius_scaling=1,max_radius_scaling=1.5, min_distance_scaling=2.5, 
            signal_window_velocity=None,signal_extent_scaling=3,max_slowness=0.5,slowness_step=0.005,velocity_perturbation=0.4,
            trace_start_time=0,taper_length_scaling=5,azimuth_step=6,min_baz_measurements=3,min_good_bazbin=5,min_beam_sharpness=0,
            doublesided=True, cc_comp ='ZZ', show_fig=True, plot_moveout=True, moveout_scaling = 4, plot_clustermap=True, 
            map_region=None,map_engine='cartopy',receiver_box=None,plot_beampower=True, plot_station_result=True, 
            verbose=False,map_region_precision=0):
    """
    Perform BANX method to compute azimuthal anisotropy.
    ===PARAMETER===
    reference_site: reference station or master station used to search for the station cluster. Contains name.
    period_band: period band for the analysis. Unit: seconds.
    reference_velocity: reference phase velocity. Unit: km/s.
    datadir: directory containing the cross-correlation data. The data should be in asdf format named as 'source_receiver*.h5'.
    outdir_root: root directory to save the results. Default is current directory.
    sampling_rate: target sampling rate. Needs to be integer times (above or below) the data sampling rate to avoid resampling error. 
                    The data will be resampled to sampling_rate (samples per second).
    min_stations: minimum number of stations in the cluster. Default is 10.
    min_snr: minimum SNR for the data. Default is 5.
    min_radius_scaling: minimum radius scaling factor. Default is 1.
    max_radius_scaling: maximum radius scaling factor. Default is 1.5.
    min_distance_scaling: minimum distance scaling factor. Default is 2.5.
    signal_extent_scaling: scaling factor for the signal extent (before and after the predicted arrival). Default is 3. (3 times the longest period)
    signal_window_velocity: group velocity to calculate the predicted arrival time to decide the signal window. 
                    If None [default], will use 80% of the reference_velocity.

    max_slowness: maximum slowness. Default is 0.5.
    slowness_step: slowness step. Default is 0.005.
    velocity_perturbation: velocity perturbation. Default is 0.4. (40%)
    trace_start_time: start time of the xcorr data. Default is 0.
    taper_length_scaling: taper length scaling factor relative to the longest period. Default is 5.

    azimuth_step: azimuthal step used in QC. Default is 6.
    min_baz_measurements: minimum number of measurements in each azimuthal bin used in QC. Should be >=3. Default is 3.
    min_good_bazbin: minimum number of good bins with >= min_baz_measurements in each azimuthal bin used in QC. Should be >=5 (recommended). Default is 5.
    min_beam_sharpness: minimum beam sharpness to pass the QC. Default is 0 [no QC by sharpness].
    doublesided: data contains both negative and positive sides of the cross-correlation data. Default is True.
    cc_comp: cross-correlation component. Default is 'ZZ'.
        
    show_fig: show figures. Default is True. Figures will be plotted if the switches are on.
    plot_moveout: plot moveout of the cluster traces to the source. Default is True.
    moveout_scaling: scaling factor for the moveout plot. Default is 4.
    plot_clustermap: plot map view of the cluster stations and the source. Default is True.
    map_engine: map plotting engine. 'cartopy' [default] or 'gmt'.
    receiver_box: receiver box for the map view. Default is None. Format [lon_min, lon_max, lat_min, lat_max].
    plot_beampower: plot beam power after beamforming with the picked max power slowness positions/values. Default is True.
    plot_station_result: plot station results with all sources after QC. Default is True.

    verbose: print verbose information. Default is False.
    ===RETURN===
    """
    if receiver_box is not None:
        if len(receiver_box) != 4:
            raise ValueError('receiver_box should be in format [lon_min, lon_max, lat_min, lat_max]')
    #check if the reference site is in the station dictionary
    if reference_site not in stationdict_all.keys():
        raise ValueError('Reference site is not in the station dictionary.')
    #check if the period band is in the correct format
    if not isinstance(period_band, (list, np.ndarray)):
        raise ValueError('Period band should be in list or numpy array format.')
    #Create directories for this period band
    outdir=os.path.join(outdir_root,str(period_band[0])+'_'+str(period_band[1]))
    if not os.path.isdir(outdir):os.makedirs(outdir, exist_ok=True)
    figdir=os.path.join(outdir,'figures')
    if not os.path.isdir(figdir):os.makedirs(figdir, exist_ok=True)
    figdir_refsite=os.path.join(figdir,reference_site)
    if not os.path.isdir(figdir_refsite):os.makedirs(figdir_refsite, exist_ok=True)

    if doublesided:
        N_Sides = 2
    else:
        N_Sides = 1

    #
    if signal_window_velocity is None:
        signal_window_velocity = 0.8*reference_velocity
    #
    SourceList_Sites = list(stationdict_all.keys())
    if sampling_rate is None:
        resample = False
    else:
        resample = True
        SamplingRate_Target = sampling_rate
        Sampling_Delta_Target = 1/SamplingRate_Target

    # Period band
    if isinstance(period_band, list):
        period_band = np.array(period_band) #seconds.
    
    # Reference period:
    if verbose: print('Reference period: ',period_band)
    Min_Period = np.min(period_band)
    Max_Period = np.max(period_band)
    # Distance constraints:
    Min_Radius = min_radius_scaling * (Min_Period * reference_velocity)
    Max_Radius = max_radius_scaling * (Max_Period * reference_velocity)
    Min_Distance = min_distance_scaling * (Max_Period * reference_velocity)

    # get source and receiver cluster information/list
    Ref_Coord = stationdict_all[reference_site] #coordinates: lat, lon in order.
    
    ReceiverCluster_Sites = [] #receivers within a certain distance from the reference site
    #compute distance from the reference site to all other sites
    ReceiverCluster_Coord = []
    SourceList_Coord =[]
    for j in range(len(SourceList_Sites)):
        SourceList_Coord.append(stationdict_all[SourceList_Sites[j]])
        dist = obspy.geodetics.base.gps2dist_azimuth(Ref_Coord[0],Ref_Coord[1],
                                                    stationdict_all[SourceList_Sites[j]][0],
                                                    stationdict_all[SourceList_Sites[j]][1])[0]/1000
        if dist <= Max_Radius:
            ReceiverCluster_Sites.append(SourceList_Sites[j])
            ReceiverCluster_Coord.append(stationdict_all[SourceList_Sites[j]])
    #
    #check for minimum number of stations
    
    if len(ReceiverCluster_Sites) < min_stations:
        return None, None
    #use sites other than the receiver cluster as the source sites.
    Source_Sites = sorted(list(set(SourceList_Sites) - set(ReceiverCluster_Sites)))
    
    print(reference_site+' -> '+str(len(ReceiverCluster_Sites))+' stations in the cluster -> '+str(len(Source_Sites))+' sources')

    #######################################
    #### Derive parameters and constants ##
    #######################################
    # Data Processing:
    AZIBIN_EDGES = np.arange(0, 180+0.5*azimuth_step, azimuth_step) #azimuthal bin egesm used in QC step.

    # Search range of local phase velocities
    Phase_Velocity_Limits = [round(reference_velocity * (1 - velocity_perturbation),2),
                            round(reference_velocity * (1 + velocity_perturbation),2)]
    print('Phase Velocity Range: ',Phase_Velocity_Limits)

    # Beamforming space:
    Slowness_Space = [-np.floor(max_slowness / slowness_step) * slowness_step, np.floor(max_slowness / slowness_step) * slowness_step];       
    Ux = np.arange(Slowness_Space[0],Slowness_Space[1]+0.5*slowness_step,slowness_step)
    Uy = np.arange(Slowness_Space[0],Slowness_Space[1]+0.5*slowness_step,slowness_step)
    # [UxUx, UyUy] = np.meshgrid(Ux, Uy)
    SlownessSamples = len(Ux)

    # gmt plotting parameters
    source_coord_array=np.array(SourceList_Coord)
    lon_all,lat_all=source_coord_array[:,1],source_coord_array[:,0]

    if receiver_box is not None:
        lon_box=[receiver_box[0],receiver_box[1],receiver_box[1],receiver_box[0],receiver_box[0]]
        lat_box=[receiver_box[2],receiver_box[2],receiver_box[3],receiver_box[3],receiver_box[2]]
    if map_region is None:
        map_region = np.round([np.min(lon_all),np.max(lon_all),np.min(lat_all),np.max(lat_all)],map_region_precision)

    if plot_clustermap:
        if map_engine.lower() == 'gmt':
            marker_style="i0.17c"
            map_style="plain"
            frame="af"
            GMT_FONT_TITLE="14p,Helvetica-Bold"
            projection = 'M3.5i'
        elif map_engine.lower() == 'cartopy':
            projection=ccrs.LambertAzimuthalEqualArea(central_longitude=np.mean([np.min(lon_all),np.max(lon_all)]), 
                                    central_latitude=np.mean([np.min(lat_all),np.max(lat_all)]), 
                                    false_easting=0.0, false_northing=0.0,globe=None)
            scale = '10m'
            states10 = cfeature.NaturalEarthFeature(
                        category='cultural',
                        name='admin_1_states_provinces_lines',
                        scale=scale,
                        facecolor='none',
                        edgecolor='k')
            country10 = cfeature.NaturalEarthFeature(
                        category='cultural',
                        name = 'admin_0_boundary_lines_land',
                        scale = scale,
                        facecolor='none',
                        edgecolor='k')
        else:
            raise ValueError('Wrong map_engine. Should be cartopy or gmt')
    ################################
    ##### Main beamforming block ###
    ################################

    """
    START BEAMFORMING LOOP
    """
    if verbose: print('  --> Beamforming loop')
    Beam_Local=[]

    for j in range(len(Source_Sites)):
        Source_Name = Source_Sites[j]
        Source_Coord = stationdict_all[Source_Name]
        flist_all=[]
        for jr in range(len(ReceiverCluster_Sites)):
            flist=[] 
            dir_temp1=os.path.join(datadir,Source_Sites[j])
            if os.path.isdir(dir_temp1):
                flist=utils.get_filelist(dir_temp1,extension='h5',pattern=ReceiverCluster_Sites[jr])
            
            dir_temp2=os.path.join(datadir,ReceiverCluster_Sites[jr])
            flist2=[]
            if os.path.isdir(dir_temp2):
                flist2=utils.get_filelist(dir_temp2,extension='h5',pattern=Source_Sites[j])
            #merge two lists
            flist.extend(flist2)
            #
            flist_all.extend(flist)
        #
        if len(flist_all) < N_Sides*min_stations:
            continue
            
        print('  Source: '+Source_Sites[j]+' -> '+str(int(len(flist_all)/N_Sides))+' pairs')

        #######################################################
        # subset receiver cluster sites to find only those with data.
        # Assemble all correlation data for the receiver cluster.
        #######################################################
        site_temp=[]
        coord_temp=[]
        ReceiverCluster_CorrData =[]
        ReceiverCluster_Dist = [] #stores distance from each station in the cluster to the source.
        for jr in range(len(ReceiverCluster_Sites)):
            sitename=ReceiverCluster_Sites[jr]
            filtered_list = [item for item in flist_all if sitename in item]
            if len(filtered_list)==N_Sides: #
                site_temp.append(sitename)
                coord_temp.append(ReceiverCluster_Coord[jr])

                cdata_temp1 = noise.extract_corrdata(filtered_list[0])
                cdata1 = cdata_temp1[list(cdata_temp1.keys())[0]][cc_comp] 
                if N_Sides == 2: #average two sides
                    cdata_temp2 = noise.extract_corrdata(filtered_list[1])
                    cdata2 = cdata_temp2[list(cdata_temp2.keys())[0]][cc_comp] 
                    if len(cdata1.data) > 0 and len(cdata2.data) > 0:
                        #average the negative and positive sides.
                        cdata1.data = (cdata1.data + cdata2.data)/2
                        cdata1.side = None
                    elif len(cdata2.data) > 1:
                        cdata1.data = cdata2.data
                    elif len(cdata1.data) > 1:
                        pass
                    else:
                        continue

                #filter the data
                cdata1.filter(fmin=1/period_band[1],fmax=1/period_band[0],corners=4,zerophase=True)
                #normalize the amplitudes.
                cdata1.data = cdata1.data/np.max(np.abs(cdata1.data))
                
                #print(len(cdata1.data))
                #collect the corrdata object.
                ReceiverCluster_CorrData.append(cdata1)
                ReceiverCluster_Dist.append(cdata1.dist)
                
        #
        if len(ReceiverCluster_CorrData) < min_stations:
            print('  Not enough stations in the cluster. Skip!')
            continue
        ReceiverCluster_Sites = site_temp
        ReceiverCluster_Coord = coord_temp

        # average positive and negative lags when reading data
        ReceiverCluster_Radius,ReceiverCluster_Center, ReceiverCluster_UTM = get_ArrayAttributes(np.array(ReceiverCluster_Coord)[:,0],
                                                                            np.array(ReceiverCluster_Coord)[:,1],get_utm=True)
        Dist2Source,_,ReceiverCluster_BAZ=obspy.geodetics.base.gps2dist_azimuth(\
                                                ReceiverCluster_Center[0],ReceiverCluster_Center[1],
                                                Source_Coord[0],Source_Coord[1])
        Dist2Source = Dist2Source/1000
        if verbose:
            print('  Cluster radius,center,dist2source,baz2source: \n    ',
              str(ReceiverCluster_Radius),str(ReceiverCluster_Center),
              str(Dist2Source),str(ReceiverCluster_BAZ))
        if ReceiverCluster_Radius < Min_Radius or Dist2Source < Min_Distance:
            print('SKIP: small cluster or too close to the source!')
            continue
        #

        """
        # Calculating the theoretical arrival times
        """
        ReceiverCluster_TravelTimes = np.array(ReceiverCluster_Dist) / signal_window_velocity
        # print(ReceiverCluster_TravelTimes)

        # Interpolate or resample if the target sampling rate is different from the data sampling rate
        Sampling_Delta_Data = ReceiverCluster_CorrData[0].dt
        Sampling_Rate_Data = int(1/Sampling_Delta_Data)
        if SamplingRate_Target == Sampling_Rate_Data:
            resample = False
        if resample:
            #max time is the maximum travel time + taper length
            MaxTime = np.round(np.max(ReceiverCluster_TravelTimes) + (signal_extent_scaling+taper_length_scaling) * Max_Period,2)
            if MaxTime > ReceiverCluster_CorrData[0].lag:
                MaxTime = ReceiverCluster_CorrData[0].lag
            # Creating a homogeneous time vector:
            TimeVector = np.arange(trace_start_time , MaxTime + 0.5*Sampling_Delta_Target, Sampling_Delta_Target)
            if (SamplingRate_Target > Sampling_Rate_Data and np.remainder(SamplingRate_Target,Sampling_Rate_Data) > 0 ) or \
                (SamplingRate_Target < Sampling_Rate_Data and np.remainder(Sampling_Rate_Data,SamplingRate_Target) > 0):
                raise ValueError('Resampling fraction (target/data) has to be an integer to avoid errors.')
            if np.abs(Sampling_Rate_Data - SamplingRate_Target) > 1:
                if verbose: print('  Resampling to (sps): ',SamplingRate_Target)
                TimeVectorData = np.arange(trace_start_time , ReceiverCluster_CorrData[0].lag + 0.5*Sampling_Delta_Data, 
                                        Sampling_Delta_Data)
                for cdata in ReceiverCluster_CorrData:
                    cdata.dt = Sampling_Delta_Target #replace the sampling interval with the target value
                    cdata.data = np.interp(TimeVector,TimeVectorData,cdata.data)
                    cdata.lag = MaxTime
        else: #keep the data sampling rate
            SamplingRate_Target = Sampling_Rate_Data
            Sampling_Delta_Target = 1/SamplingRate_Target
            #max time is the maximum travel time + taper length
            MaxTime = np.round(np.max(ReceiverCluster_TravelTimes) + (taper_length_scaling * Max_Period),2)
            # set the max time to the nearest multiple of the data sampling interval.
            if np.remainder(MaxTime,Sampling_Delta_Data) > 0:
                MaxTime = np.round(MaxTime/Sampling_Delta_Data)*Sampling_Delta_Data
            if MaxTime > ReceiverCluster_CorrData[0].lag:
                MaxTime = ReceiverCluster_CorrData[0].lag
            TimeVector = np.arange(trace_start_time , MaxTime + 0.5*Sampling_Delta_Target, Sampling_Delta_Target)

            # cut the data to the max time in corrdata.
            for cdata in ReceiverCluster_CorrData:
                cdata.data = cdata.data[:len(TimeVector)]
                cdata.lag = MaxTime
                
        # Quality control with SNR cutoff
        #1. set a time window around the predicted travel time.
        #2. get Max of the signal.
        #3. Mute the signal and get the rms of the noise outside the signal window
        #4. Compute SNR and check for QC
        GoodMatrix=[]
        GoodDist=[]
        GoodCoord=[]
        x_km=[]
        y_km=[]
        # UTM demean
        x_km_all = (ReceiverCluster_UTM[0] - np.min(ReceiverCluster_UTM[0]))/1000
        y_km_all = (ReceiverCluster_UTM[1] - np.min(ReceiverCluster_UTM[1]))/1000
        if plot_moveout:
            plt.figure(figsize=[5,4])
        for ic,cdata in enumerate(ReceiverCluster_CorrData):
            data_temp=cdata.data.copy()
            tt = ReceiverCluster_TravelTimes[ic]
            #signal time window is half longest period away from the predicted time.
            signal_twin = [tt - signal_extent_scaling*np.max(period_band), tt + signal_extent_scaling*np.max(period_band)]
            if signal_twin[0] < trace_start_time: signal_twin[0]=trace_start_time
            if signal_twin[1] > MaxTime: signal_twin[1]=MaxTime
            signal_idx = [np.argmin(np.abs(TimeVector - signal_twin[0])), np.argmin(np.abs(TimeVector - signal_twin[1]))]
            signal_absmax = np.max(np.abs(data_temp[signal_idx[0]:signal_idx[1]]))
            
            #mute signal window
            data_temp[signal_idx[0]:signal_idx[1]]=0.0
            noise_absrms = utils.rms(data_temp)
            snr = signal_absmax / noise_absrms
            # print('    SNR: ',snr)
            if snr >= min_snr and cdata.dist >= Min_Distance:
                data_mutenoise=cdata.data.copy()
                ################################################################################
                # !!!!!!!!!!!!!!!! mute noise window to zeros !!!!!!!!!!!!!!!!
                ################################################################################
                data_mutenoise[:signal_idx[0]]=0.0
                data_mutenoise[signal_idx[1]:]=0.0
                
                GoodMatrix.append(data_mutenoise)
                GoodDist.append(cdata.dist)
                GoodCoord.append(ReceiverCluster_Coord[ic])
                #x and y coordinates of the stations
                x_km.append(x_km_all[ic])
                y_km.append(y_km_all[ic])
                
                if plot_moveout:
                    plt.plot(TimeVector,moveout_scaling * cdata.data + cdata.dist,'k-',lw=0.5)
                    plt.plot(tt,cdata.dist,'r|',markerfacecolor='none',markersize=5)
                    plt.plot([signal_twin[0],signal_twin[1]],[cdata.dist,cdata.dist],'b|',lw=1)
            #
        #
        
        x_km=np.array(x_km)
        y_km=np.array(y_km)
        if len(GoodMatrix) < min_stations:
            print('  Not enough stations after SNR cutoff. Skip!')
            plt.close()
            continue
        if plot_moveout:
            plt.xlim([trace_start_time,MaxTime])
            plt.ylim([np.min(GoodDist) - 5*moveout_scaling, np.max(GoodDist) + 5*moveout_scaling])
            plt.xlabel('Time (s)')
            plt.ylabel('Distance (km)')
            plt.title('Ref: '+reference_site+', Source: '+Source_Name+': '+str(len(GoodDist)))
            plt.savefig(os.path.join(figdir_refsite,reference_site+'_'+Source_Name+'_moveout.pdf'))
            if show_fig: plt.show()
            else: plt.close()
        GoodMatrix = np.array(GoodMatrix)
        GoodCoord = np.array(GoodCoord)
        GoodDist = np.array(GoodDist)
        if verbose: print('  Num. good trace: ',GoodMatrix.shape[0])

        #free some memory
        del ReceiverCluster_CorrData
        #
        # Plot map view of good stations and the source.
        if plot_clustermap:
            if map_engine.lower() == 'cartopy':
                plt.figure(figsize=[5,4],facecolor='w')
                ax=plt.axes(projection=projection)
                #plot receiver cluster
                plt.plot(GoodCoord[:,1],GoodCoord[:,0],'r^',markersize=4,markerfacecolor='none',transform=ccrs.PlateCarree())
                #plot receiver box
                if receiver_box is not None:
                    plt.plot(lon_box,lat_box,'b-',lw=1,transform=ccrs.PlateCarree())
                #plot source
                plt.plot(Source_Coord[1],Source_Coord[0],'r*',markersize=10, markerfacecolor='none',transform=ccrs.PlateCarree())
                #plot line from cluster center to source
                plt.plot([ReceiverCluster_Center[1],Source_Coord[1]],[ReceiverCluster_Center[0],Source_Coord[0]],'k-',
                        lw=1.5,transform=ccrs.PlateCarree())
                ax.set_extent((map_region[0],map_region[1],map_region[2]*0.95,map_region[3]*1.0))
                ax.coastlines(resolution='10m',color='k', linewidth=.5)
                gl=ax.gridlines(crs=ccrs.PlateCarree(),draw_labels=True, dms=True, x_inline=False, y_inline=False,
                                color='gray', alpha=0.5, linestyle='--',linewidth=0.5)
                ax.add_feature(country10,lw=1)
                ax.add_feature(states10,lw=0.5)
                gl.xlocator = mticker.FixedLocator(np.arange(map_region[0],map_region[1]+2.5,5))
                gl.ylocator = mticker.FixedLocator(np.arange(map_region[2],map_region[3]+2.5,5))
                gl.xlabels_top = False
                plt.title(reference_site+' -> '+Source_Name,pad=20)
                
                plt.savefig(os.path.join(figdir_refsite,reference_site+'_'+Source_Name+'_stationmap.pdf'),format='pdf',dpi=300)
                if show_fig: plt.show()
                else: plt.close()
            elif map_engine.lower() == 'gmt':
                if verbose: print('  Plotting cluster map ...')
                fig = gmt.Figure()
                gmt.config(MAP_FRAME_TYPE=map_style, FONT_TITLE=GMT_FONT_TITLE)
                fig.coast(region=map_region, resolution="f",projection=map_projection, 
                        water="0/180/255",frame=frame,land="240",borders=["1/1p,black", "2/0.5p,100"])
                fig.basemap(frame='+t'+reference_site+' -> '+Source_Name)
                #plot receiver cluster
                fig.plot(x=GoodCoord[:,1],y=GoodCoord[:,0],style=marker_style,pen="0.5p,red",)
                #plot receiver box
                if receiver_box is not None:
                    fig.plot(x=lon_box,y=lat_box,pen="1p,blue",)
                #plot source
                fig.plot(x=Source_Coord[1],y=Source_Coord[0],style="a0.3c",pen="1p,red",)
                #plot line from cluster center to source
                fig.plot(x=[ReceiverCluster_Center[1],Source_Coord[1]],
                         y=[ReceiverCluster_Center[0],Source_Coord[0]],
                         pen="1p,black",)
                fig.savefig(os.path.join(figdir_refsite,reference_site+'_'+Source_Name+'_stationmap.pdf'))
                if show_fig: fig.show()
                else: gmt.set_display('none')
       
        # Beamforming
        if verbose: print('  Performing beamforming ...')
        Slowness_Image = np.zeros((SlownessSamples, SlownessSamples))
        Slowness_Image.fill(np.nan)
        for k in range(Slowness_Image.shape[0]):
            for kk in range(Slowness_Image.shape[1]):
                # Backazimuth and velocity:
                Slowness_Temp = utils.cart2compass(Ux[k], Uy[kk])[1]
                Velocity_Temp = 1 / Slowness_Temp
                
                if Velocity_Temp < Phase_Velocity_Limits[0] or Velocity_Temp > Phase_Velocity_Limits[1]: 
                    continue
                #
                # Time shifts:
                TimeShifts = x_km * Ux[k] + y_km * Uy[kk]
                SampleShifts = np.round(TimeShifts * SamplingRate_Target)

                Shifted_Traces = np.zeros(GoodMatrix.shape)
                for l in range(GoodMatrix.shape[0]):
                    shift = int(SampleShifts[l])
                    Trace_Temp = np.roll(GoodMatrix[l], shift)
                    
                    if shift < 0:
                        Trace_Temp[-shift:] = 0
                    elif shift > 0:
                        Trace_Temp[1:shift] = 0

                    # Allocating the shifted traces:
                    Shifted_Traces[l] = Trace_Temp
                #
                # Measuring power
                Beam = np.sum(Shifted_Traces, axis=0) / Shifted_Traces.shape[0]
                Slowness_Image[kk,k] = np.sum(np.power(Beam,2))
        ##
        if verbose: print('  ... Done beamforming')

        #Free some memory
        del GoodMatrix
        
        # Extracting the velocity and azimuth:
        Power = np.nanmax(Slowness_Image)
        Max_idx=np.nanargmax(Slowness_Image)
        Max_idx_sub=np.unravel_index(Max_idx,Slowness_Image.shape)
        
        Best_Ux = Ux[Max_idx_sub[1]]
        Best_Uy = Uy[Max_idx_sub[0]]
        Beam_BAZ, Beam_U = utils.cart2compass(Best_Ux, Best_Uy)
        Beam_Velocity = 1 / Beam_U
        Beam_Sharpness = np.nanmax(Slowness_Image) / np.nanmedian(Slowness_Image) #sharpness of the beam image.
        if verbose: print('  Beam results (Max power, sharpness, Best Ux, Best Uy, BAZ, velocity): %.4f, %.2f, %.4f, %.4f, %.4f, %.4f'
              %(Power,Beam_Sharpness,Best_Ux,Best_Uy,Beam_BAZ,Beam_Velocity))
        # Allocating the output:
        Beam_Local_temp = [ReceiverCluster_Center[0],
                           ReceiverCluster_Center[1],
                           Beam_Sharpness,
                           Beam_Velocity,
                           Beam_BAZ,
                           Power,
                           ReceiverCluster_Radius,
                           len(GoodDist),
                           reference_velocity,
                           ReceiverCluster_BAZ]
        #
        if np.isnan(np.array(Beam_Local_temp)).any():
            continue
        else:
            Beam_Local.append(Beam_Local_temp)
        #
        if plot_beampower:
            plt.figure(figsize=[5,4])
            plt.imshow(Slowness_Image/Power,extent=[Ux[0], Ux[-1], Uy[-1], Uy[0]],vmin=0,vmax=1)
            plt.plot(Best_Ux,Best_Uy,'ro',markerfacecolor='none')
            plt.xlim([Ux[0], Ux[-1]])
            plt.ylim([Uy[0], Uy[-1]])
            plt.xlabel('Ux')
            plt.ylabel('Uy')
            plt.colorbar(label='Normalized Power')
            plt.title('Ref: '+reference_site+', Source: '+Source_Name+'\nSharpness: %.1f, Velocity: %.2f km/s'%(Beam_Sharpness,Beam_Velocity))
            plt.savefig(os.path.join(figdir_refsite,reference_site+'_'+Source_Name+'_beam.pdf'))
            if show_fig: plt.show()
            else: plt.close()

        #
        del Slowness_Image
    # End of loop for one reference site to all available sources.
    if len(Beam_Local) < 1:
        return None, None
    #
    
    Beam_Local = pd.DataFrame(np.array(Beam_Local),columns=['lat','lon','sharpness','velocity','baz','power',
                                                            'radius','num','velref','baz_cluster'])
    if verbose: print('  Fitting anisotropy parameters ...')
    #reuse good_idx to filter the data
    good_idx = np.where((Beam_Local['sharpness'] >= min_beam_sharpness) & (Beam_Local['velocity'] >= Phase_Velocity_Limits[0]) & \
                        (Beam_Local['velocity'] <= Phase_Velocity_Limits[1]))[0]
    Beam_Local_subset = Beam_Local.iloc[good_idx]
    
    # following are needed for QC of the BAZ coverage.
    BAZ_check = Beam_Local_subset['baz'].copy()
    BAZ_check[BAZ_check<0] += 180
    BAZ_check[BAZ_check>180] -= 180
    histcount=np.histogram(BAZ_check,bins=AZIBIN_EDGES)[0]
    goodazbin_idx = np.where(histcount >= min_baz_measurements)[0]
    if len(goodazbin_idx) < min_good_bazbin:
        print('  Not enough azimuthal coverage of sources after QC with sharpness and velocity value. Skip!')
        return None, None
    
    # Estimate anisotropy parameters by curve fitting.
    # Only use the bins with good measurements. Need to add weight in the future.
    fitcoef=curve_fit(compute_anisotropy,Beam_Local_subset['baz'],Beam_Local_subset['velocity'],p0=[3.5,0.1,0.1])[0]
    A0 = fitcoef[0]
    A1 = fitcoef[1]
    A2 = fitcoef[2]

    # Converting the anisotropy to percentage:
    RHO = np.sqrt(A1**2 + A2**2) / A0 * 100
    THETA = np.degrees(0.5 * np.arctan2(A2, A1))
    if THETA < 0:
        THETA += 180
    elif THETA >= 0 and THETA < 90:
        THETA = np.abs(THETA - 90)
    elif THETA >= 90 and THETA <= 180:
        THETA = np.abs(180 - THETA)
    else:
        pass
    print('  Anisotropy: %.2f%%, Fast direction (degrees from north): %.2f'%(RHO,THETA))
    ReceiverCluster_Center_Final = np.round(get_ArrayAttributes(Beam_Local_subset['lat'].values,Beam_Local_subset['lon'].values)[1],4)

    if verbose: print('  Final cluster center: %.4f, %.4f'%(ReceiverCluster_Center_Final[0],ReceiverCluster_Center_Final[1]))

    # Save the anisotropy results
    beam_outfile=os.path.join(outdir,reference_site+'_beam.csv')
    Beam_Local_subset.to_csv(beam_outfile,index=False)
    print('  Beam results saved to: ',beam_outfile)

    anisotropy=[period_band[0],period_band[1],Max_Radius,min_stations,
                ReceiverCluster_Center_Final[0],ReceiverCluster_Center_Final[1],
                os.path.split(beam_outfile)[1],A0,A1,A2,RHO,THETA]
    anisotropy_outfile=os.path.join(outdir,reference_site+'_anisotropy.csv')
    anisotropy = pd.DataFrame([anisotropy],columns=['Period_min','Period_max','Max_Radius','min_stations',
                                                    'lat','lon','beamfile','A0','A1','A2','RHO','THETA'])
    anisotropy.to_csv(anisotropy_outfile,index=False)
    print('  Anisotropy results saved to: ',anisotropy_outfile)

    # plot phase velocity of the reference station
    if plot_station_result:
        fig=plt.figure(figsize=[7,5])
        ax = fig.add_subplot(6,1,(1,2))
        ax.hist(Beam_Local_subset['baz'],bins=180,range=[0,360],color='tab:blue',label='BAZ')
        # plt.xlabel('BAZ')
        ax.set_ylabel('Source count')
        ax.set_xlim([0,360])
        ax.grid(lw=0.5)
        ax.set_title('Ref site: '+reference_site)

        ax2 = fig.add_subplot(6,1,(3,6))
        #plt beam baz v.s. velocity
        sc=ax2.scatter(Beam_Local_subset['baz'],Beam_Local_subset['velocity'],s=5*Beam_Local_subset['sharpness'],marker='o',linewidths=0.5,
                    facecolor='none',c=Beam_Local_subset['sharpness'],edgecolor='k',label='Data')
        #plot the fitted curve
        x_fit=np.linspace(0,360,100)
        y_fit=compute_anisotropy(x_fit,*fitcoef)
        ax2.plot(x_fit,y_fit,'r-',label='Fitted curve')
        ax2.plot(x_fit,np.ones_like(x_fit)*fitcoef[0],'k--',label='Isotropic velocity')
        ax2.set_xlabel('BAZ')
        ax2.set_ylabel('Velocity')
        # plt.title('Ref site: '+reference_site+')
        ax2.set_xlim([0,360])
        ax2.set_ylim([Phase_Velocity_Limits[0],Phase_Velocity_Limits[1]])
        ax2.grid(lw=0.5)
        ax2.legend(ncol=3,loc='upper center')
        fig.colorbar(sc,label='Beam sharpness',location='bottom',pad=0.2)
        fig.tight_layout()
        plt.savefig(os.path.join(figdir_refsite,reference_site+'_station_result.pdf'))
        if show_fig: plt.show()
        else: plt.close()
#
    return Beam_Local, anisotropy
######################### End of BANX method ##############################


